from mitmproxy.test import tflow, tutils

from hippolyzer.lib.proxy.caps import CapType
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.message_logger import HTTPMessageLogEntry
from hippolyzer.lib.proxy.test_utils import BaseProxyTest


class TestHTTPFlows(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.region = self.session.register_region(
            ("127.0.0.1", 2),
            "https://test.localhost:4/foo",
            handle=90,
        )
        self.region.update_caps({
            "FakeCap": "http://example.com",
            "ViewerAsset": "http://assets.example.com",
        })

    async def test_request_formatting(self):
        req = tutils.treq(host="example.com", port=80)
        fake_flow = tflow.tflow(req=req, resp=tutils.tresp())
        flow = HippoHTTPFlow.from_state(fake_flow.get_state(), self.session_manager)
        # Make sure cap resolution works correctly
        flow.cap_data = self.session_manager.resolve_cap(flow.request.url)
        entry = HTTPMessageLogEntry(flow)
        self.assertEqual(entry.request(beautify=True), """GET [[FakeCap]]/path HTTP/1.1\r
# http://example.com/path\r
header: qvalue\r
content-length: 7\r
\r
content""")

    async def test_binary_request_formatting(self):
        req = tutils.treq(host="example.com", port=80)
        fake_flow = tflow.tflow(req=req, resp=tutils.tresp())
        flow = HippoHTTPFlow.from_state(fake_flow.get_state(), self.session_manager)
        # This should trigger the escaped body path without changing content-length
        flow.request.content = b"c\x00ntent"
        entry = HTTPMessageLogEntry(flow)
        self.assertEqual(entry.request(beautify=True), """GET http://example.com/path HTTP/1.1\r
header: qvalue\r
content-length: 7\r
X-Hippo-Escaped-Body: 1\r
\r
c\\x00ntent""")

    async def test_llsd_response_formatting(self):
        fake_flow = tflow.tflow(req=tutils.treq(), resp=tutils.tresp())
        flow = HippoHTTPFlow.from_state(fake_flow.get_state(), self.session_manager)
        # Half the time LLSD is sent with a random Content-Type and no PI indicating
        # what flavor of LLSD it is. Make sure the sniffing works correctly.
        flow.response.content = b"<llsd><integer>1</integer></llsd>"
        entry = HTTPMessageLogEntry(flow)
        self.assertEqual(entry.response(beautify=True), """HTTP/1.1 200 OK\r
header-response: svalue\r
content-length: 33\r
\r
<?xml version="1.0" ?>
<llsd>
<integer>1</integer>
</llsd>
""")

    async def test_flow_state_serde(self):
        fake_flow = tflow.tflow(req=tutils.treq(host="example.com"), resp=tutils.tresp())
        flow = HippoHTTPFlow.from_state(fake_flow.get_state(), self.session_manager)
        # Make sure cap resolution works correctly
        flow.cap_data = self.session_manager.resolve_cap(flow.request.url)
        flow_state = flow.get_state()
        new_flow = HippoHTTPFlow.from_state(flow_state, self.session_manager)
        self.assertIs(self.session, new_flow.cap_data.session())

    async def test_http_asset_repo(self):
        asset_repo = self.session_manager.asset_repo
        asset_id = asset_repo.create_asset(b"foobar", one_shot=True)
        req = tutils.treq(host="assets.example.com", path=f"/?animatn_id={asset_id}")
        fake_flow = tflow.tflow(req=req)
        flow = HippoHTTPFlow.from_state(fake_flow.get_state(), self.session_manager)
        # Have to resolve cap data so the asset repo knows this is an asset server cap
        flow.cap_data = self.session_manager.resolve_cap(flow.request.url)
        self.assertTrue(asset_repo.try_serve_asset(flow))
        self.assertEqual(b"foobar", flow.response.content)

    async def test_temporary_cap_resolution(self):
        self.region.register_cap("TempExample", "http://not.example.com", CapType.TEMPORARY)
        self.region.register_cap("TempExample", "http://not2.example.com", CapType.TEMPORARY)
        # Resolving the cap should consume it
        cap_data = self.session_manager.resolve_cap("http://not.example.com")
        self.assertEqual(cap_data.cap_name, "TempExample")
        # A CapData object should always be returned, but the cap_name field will be None
        new_cap_data = self.session_manager.resolve_cap("http://not.example.com")
        self.assertIsNone(new_cap_data.cap_name)
        # The second temp cap with the same name should still be in there
        cap_data = self.session_manager.resolve_cap("http://not2.example.com")
        self.assertEqual(cap_data.cap_name, "TempExample")



"""
Make most object textures monochrome.

Avoids affecting materials and profile pictures and the like by
replacing textures in TEs with an alias and only mutating requests
for those alias IDs.

Demonstrates a multiprocessing / queue consumer pattern for addons that
need to do expensive, potentially blocking work in the background.

This will make your CPU fan go crazy so enjoy that.
"""

import copy
import ctypes
import multiprocessing
import queue
import secrets
import statistics
import time
import traceback

import glymur
import numpy as np
from mitmproxy.http import HTTPFlow

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.jp2_utils import BufferedJp2k
from hippolyzer.lib.base.multiprocessing_utils import ParentProcessWatcher
from hippolyzer.lib.base.templates import TextureEntryCollection
from hippolyzer.lib.proxy.addon_utils import AssetAliasTracker, BaseAddon, GlobalProperty, AddonProcess
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


glymur.set_option('lib.num_threads', 4)

# These should never be replaced, they're only used as aliases to tell the viewer
# it should fetch the relevant texture from the appearance service
BAKES_ON_MESH_TEXTURE_IDS = {UUID(x) for x in (
    "5a9f4a74-30f2-821c-b88d-70499d3e7183",
    "ae2de45c-d252-50b8-5c6e-19f39ce79317",
    "24daea5f-0539-cfcf-047f-fbc40b2786ba",
    "52cc6bb6-2ee5-e632-d3ad-50197b1dcb8a",
    "43529ce8-7faa-ad92-165a-bc4078371687",
    "09aac1fb-6bce-0bee-7d44-caac6dbb6c63",
    "ff62763f-d60a-9855-890b-0c96f8f8cd98",
    "8e915e25-31d1-cc95-ae08-d58a47488251",
    "9742065b-19b5-297c-858a-29711d539043",
    "03642e83-2bd1-4eb9-34b4-4c47ed586d2d",
    "edd51b77-fc10-ce7a-4b3d-011dfc349e4f",
)}


def _modify_crc(crc_tweak: int, crc_val: int):
    return ctypes.c_uint32(crc_val ^ crc_tweak).value


class MonochromeAddon(BaseAddon):
    NUM_CONSUMERS = 6
    mono_crc_xor: int = GlobalProperty()
    mono_tracker: AssetAliasTracker = GlobalProperty(AssetAliasTracker)

    def __init__(self):
        # These are global and should die whenever the addon reloads,
        # so they can live on the instance rather than in addon context
        self.mono_addon_shutdown_signal = multiprocessing.Event()
        self.image_resp_queue = multiprocessing.Queue()

    def handle_init(self, session_manager: SessionManager):
        to_proxy_queue = session_manager.flow_context.to_proxy_queue
        for _ in range(self.NUM_CONSUMERS):
            # We must use AddonProcess rather than multiprocessing.Process because our
            # target function is in a dynamically loaded addon. See AddonProcess' docstring.
            AddonProcess(
                target=_process_image_queue,
                args=(self.mono_addon_shutdown_signal, self.image_resp_queue, to_proxy_queue),
                daemon=True,
            ).start()

        # We want CRCs that are stable for the duration of the addon's life, but will
        # cause a cache miss for objects cached before. Generate a random value
        # to XOR all CRCs with.
        self.mono_crc_xor = secrets.randbits(32)
        self.mono_tracker.invalidate_aliases()

    def handle_session_init(self, session: Session):
        # We loaded while this session was active, re-request all objects we
        # know about so we can process them
        if session.main_region:
            object_manager = session.main_region.objects
            object_manager.request_missing_objects()
            object_manager.request_objects([o.LocalID for o in object_manager.all_objects])

    def handle_unload(self, session_manager: SessionManager):
        # Tell queue consumers to shut down
        self.mono_addon_shutdown_signal.set()

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        tracker = self.mono_tracker
        if message.name == "ObjectUpdateCached":
            for block in message["ObjectData"]:
                # Cached only really has a CRC, this will force the cache miss.
                block["CRC"] = _modify_crc(self.mono_crc_xor, block["CRC"])
        elif message.name == "ObjectUpdate":
            for block in message["ObjectData"]:
                block["CRC"] = _modify_crc(self.mono_crc_xor, block["CRC"])
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue

                parsed_te = self._make_te_monochrome(tracker, parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "ImprovedTerseObjectUpdate":
            for block in message["ObjectData"]:
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue
                parsed_te = self._make_te_monochrome(tracker, parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "AvatarAppearance":
            for block in message["ObjectData"]:
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue
                # Have to hook AppearanceService for this to work.
                parsed_te = self._make_te_monochrome(tracker, parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "ObjectUpdateCompressed":
            for block in message["ObjectData"]:
                update_data = block.deserialize_var("Data")
                if not update_data:
                    continue
                update_data["CRC"] = _modify_crc(self.mono_crc_xor, update_data["CRC"])
                if not update_data.get("TextureEntry"):
                    continue

                update_data["TextureEntry"] = self._make_te_monochrome(
                    tracker,
                    update_data["TextureEntry"],
                )
                block.serialize_var("Data", update_data)
        elif message.name == "RegionHandshake":
            for field_name, val in message["RegionInfo"][0].items():
                if field_name.startswith("Terrain") and isinstance(val, UUID):
                    message["RegionInfo"][field_name] = tracker.get_alias_uuid(val)

    @staticmethod
    def _make_te_monochrome(tracker: AssetAliasTracker, parsed_te: TextureEntryCollection):
        # Need a deepcopy because TEs are owned by the ObjectManager
        # and we don't want to change the canonical view.
        parsed_te = copy.deepcopy(parsed_te)
        for k, v in parsed_te.Textures.items():
            if v in BAKES_ON_MESH_TEXTURE_IDS:
                continue
            # Replace textures with their alias to bust the viewer cache
            parsed_te.Textures[k] = tracker.get_alias_uuid(v)
        for k, v in parsed_te.Color.items():
            # Convert face colors to monochrome, keeping alpha byte the same
            avg_val = int(round(statistics.mean((v[0], v[1], v[2]))))
            parsed_te.Color[k] = bytes((avg_val,) * 3 + (v[3],))
        return parsed_te

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        cap_data = flow.cap_data
        if not cap_data:
            return
        is_appearance = cap_data.cap_name == "AppearanceService"
        if not (cap_data.asset_server_cap or is_appearance):
            return

        if is_appearance:
            # Baked layers come from the sim-local appearance service, not the asset server.
            # Its request URIs look a little different.
            texture_id = flow.request.url.split('/')[-1]
        else:
            texture_id = flow.request.query.get("texture_id")

        if not texture_id:
            return

        orig_texture_id = self.mono_tracker.get_orig_uuid(UUID(texture_id))
        if not orig_texture_id:
            return
        if orig_texture_id in BAKES_ON_MESH_TEXTURE_IDS:
            return

        # The request was for a fake texture ID we created, rewrite the request to
        # request the real asset and mark the flow for modification once we receive
        # the image.
        if is_appearance:
            split_url = flow.request.url.split('/')
            split_url[-1] = str(orig_texture_id)
            flow.request.url = '/'.join(split_url)
        else:
            flow.request.query["texture_id"] = str(orig_texture_id)

        flow.can_stream = False
        flow.metadata["make_monochrome"] = True
        # We can't parse a partial J2C. This is probably a little inefficient since we're
        # liable to have multiple in-flight requests for different ranges of a file at any
        # given time, and we'll have to re-encode multiple times. Meh.
        flow.request.headers.pop("Range", None)

    def handle_http_response(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if not flow.metadata.get("make_monochrome"):
            return

        if flow.response.status_code not in (200, 206):
            return

        # Don't send the callback to the proxy immediately, our queue consumer is
        # messing with the image data and will send the callback itself.
        flow.take()
        # Put the serialized HTTP flow on the queue for one of the consumers to pick up
        self.image_resp_queue.put(flow.get_state())
        return True


def _process_image_queue(
        shutdown_signal: multiprocessing.Event,
        image_resp_queue: multiprocessing.Queue,
        to_proxy_queue: multiprocessing.Queue
):
    watcher = ParentProcessWatcher(shutdown_signal)
    while not watcher.check_shutdown_needed():
        try:
            flow_state = image_resp_queue.get(False)
        except queue.Empty:
            # Ok to block since we're in our own process
            time.sleep(0.01)
            continue

        # Use HTTPFlow since we don't have a session manager and don't need
        # to understand any Hippolyzer-specific fields
        flow: HTTPFlow = HTTPFlow.from_state(flow_state)
        try:
            flow.response.content = _make_jp2_monochrome(flow.response.content)
            # Not a range anymore, update the headers and status.
            flow.response.headers.pop("Content-Range", None)
            flow.response.status_code = 200
        except:
            # Just log the exception and pass the image through unmodified
            traceback.print_exc()

        # Put our modified response directly on the mitmproxy response queue,
        # no point sending it back to the addon.
        to_proxy_queue.put(("callback", flow.id, flow.get_state()))


def _make_jp2_monochrome(jp2_data: bytes) -> bytes:
    j = BufferedJp2k(jp2_data)
    # Less than three components? Probably monochrome already.
    if len(j.shape) < 3 or j.shape[2] < 3:
        return jp2_data

    # Downscale if it'll be a huge image, compression is slow.
    if any(c >= 1024 for c in j.shape[0:2]):
        pixels = j[::4, ::4]
    elif any(c >= 512 for c in j.shape[0:2]):
        pixels = j[::2, ::2]
    else:
        pixels = j[::]

    for row in pixels:
        for col in row:
            # Simple average value monochrome conversion
            # Don't touch any component after the third. Those
            # have alpha and baked layer related data.
            col[0:3] = np.mean(col[0:3])

    # RGB, we can convert this to a single monochrome channel since
    # we don't have to worry about messing with alpha.
    if pixels.shape[2] == 3:
        pixels = pixels[:, :, 0:1]
        # Inform glymur of the new array shape
        j.shape = (pixels.shape[0], pixels.shape[1], 1)

    j[::] = pixels
    return bytes(j)


addons = [MonochromeAddon()]



"""
ObjectUpdate blame tracker, to figure out what objects are spamming updates

Assumes that you've received a full ObjectUpdate for everything (meaning the proxy
object tracker knows about it) and that you have received an ObjectProperties for
everything you want the name of. You can force a full ObjectUpdate for everything
by relogging with an empty object cache. Doing the "precache_objects" command
before you start tracking can help too.
"""
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.base.templates import PCode
from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message, SessionProperty
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class ObjectUpdateBlameAddon(BaseAddon):
    update_blame_counter: Counter[UUID] = SessionProperty(Counter)
    should_track_update_blame: bool = SessionProperty(False)

    @handle_command()
    async def precache_objects(self, _session: Session, region: ProxiedRegion):
        """
        Make the proxy's object tracker request any missing objects

        Should be done before tracking update blame to make sure the proxy
        knows about any objects that are cached in the client but not by the proxy
        """
        region.objects.request_missing_objects()

    @handle_command()
    async def object_cache_miss_stats(self, _session: Session, region: ProxiedRegion):
        show_message(len(region.objects.missing_locals))

    @handle_command()
    async def track_update_blame(self, _session: Session, _region: ProxiedRegion):
        self.should_track_update_blame = True

    @handle_command()
    async def untrack_update_blame(self, _session: Session, _region: ProxiedRegion):
        self.should_track_update_blame = False

    @handle_command()
    async def clear_update_blame(self, _session: Session, _region: ProxiedRegion):
        self.update_blame_counter.clear()

    @handle_command()
    async def dump_update_blame(self, _session: Session, region: ProxiedRegion):
        print("ObjectUpdate blame:")
        for obj_id, count in self.update_blame_counter.most_common(50):
            obj = region.objects.lookup_fullid(obj_id)
            name = obj.Name if obj and obj.Name else "<Unknown>"
            print(f"{obj_id} ({name!r}): {count}")

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        if not self.should_track_update_blame:
            return
        if region != session.main_region:
            return
        # Log this as related to the parent object unless the parent is an avatar
        if obj.Parent and obj.Parent.PCode != PCode.AVATAR:
            obj = obj.Parent

        if obj.PCode != PCode.PRIMITIVE:
            return

        self.update_blame_counter[obj.FullID] += 1
        # Ask the region for the object name if we don't know it
        if obj.Name is None:
            region.objects.request_object_properties(obj)


addons = [ObjectUpdateBlameAddon()]



import unittest

import numpy as np

from hippolyzer.lib.base.mesh_skeleton import load_avatar_skeleton


class TestSkeleton(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.skeleton = load_avatar_skeleton()

    def test_get_joint(self):
        node = self.skeleton["mNeck"]
        self.assertEqual("mNeck", node.name)
        self.assertEqual(self.skeleton, node.skeleton())

    def test_get_joint_index(self):
        self.assertEqual(7, self.skeleton["mNeck"].index)
        self.assertEqual(113, self.skeleton["mKneeLeft"].index)

    def test_get_joint_parent(self):
        self.assertEqual("mChest", self.skeleton["mNeck"].parent.name)

    def test_get_joint_matrix(self):
        expected_mat = np.array([
            [1., 0., 0., -0.01],
            [0., 1., 0., 0.],
            [0., 0., 1., 0.251],
            [0., 0., 0., 1.]
        ])
        np.testing.assert_equal(expected_mat, self.skeleton["mNeck"].matrix)

    def test_get_inverse_joint(self):
        self.assertEqual("R_CLAVICLE", self.skeleton["L_CLAVICLE"].inverse.name)
        self.assertEqual(None, self.skeleton["mChest"].inverse)
        self.assertEqual("mHandMiddle1Right", self.skeleton["mHandMiddle1Left"].inverse.name)
        self.assertEqual("RIGHT_HANDLE", self.skeleton["LEFT_HANDLE"].inverse.name)



import os.path
import unittest

import glymur
from glymur.codestream import CMEsegment

from hippolyzer.lib.base.jp2_utils import BufferedJp2k

BASE_PATH = os.path.dirname(os.path.abspath(__file__))


@unittest.skipIf(glymur.jp2k.opj2.OPENJP2 is None, "OpenJPEG library missing")
class TestJP2Utils(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open(os.path.join(BASE_PATH, "test_resources", "plywood.j2c"), "rb") as f:
            cls.j2c_bytes = f.read()

    def test_load_j2c(self):
        j = BufferedJp2k(contents=self.j2c_bytes)
        j.parse()
        # Last segment in the header is the comment section
        com: CMEsegment = j.codestream.segment[-1]
        self.assertEqual("CME", com.marker_id)
        # In this case the comment is the encoder version
        self.assertEqual(b'Kakadu-3.0.3', com.ccme)

    def test_read_j2c_data(self):
        j = BufferedJp2k(self.j2c_bytes)
        pixels = j[::]
        self.assertEqual((512, 512, 3), pixels.shape)

    def test_save_j2c_data(self):
        j = BufferedJp2k(self.j2c_bytes)
        pixels = j[::]
        j[::] = pixels
        new_j2c_bytes = bytes(j)
        self.assertNotEqual(self.j2c_bytes, new_j2c_bytes)
        # Glymur will have replaced the CME section with its own
        self.assertIn(b"Created by OpenJPEG", new_j2c_bytes)



"""
Message Mirror

Re-routes messages through the circuit of another agent running through this proxy,
rewriting the messages to use the credentials tied to that circuit.

Useful if you need to quickly QA authorization checks on a message handler or script.
Or if you want to chat as two people at once. Whatever.
Also shows some advanced ways of managing / rerouting Messages and HTTP flows.

Fiddle with the values of `SEND_NORMALLY` and `MIRROR` to change how and which
messages get moved to other circuits.

Usage: /524 mirror_to <mirror_agent_uuid>
To Disable: /524 mirror_to
"""

import weakref
from typing import Optional

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.template_dict import DEFAULT_TEMPLATE_DICT
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty, show_message
from hippolyzer.lib.proxy.commands import handle_command, Parameter, parse_bool
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.caps import CapData, CapType
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager

# Things that make no sense to mirror, or will make everything explode if mirrored.
SEND_NORMALLY = {
    'StartPingCheck', 'CompletePingCheck', 'PacketAck', 'SimulatorViewerTimeMessage', 'SimStats',
    'SoundTrigger', 'EventQueueGet', 'GetMesh', 'GetMesh2', 'ParcelDwellRequest', 'ViewerEffect', 'ViewerStats',
    'ParcelAccessListRequest', 'FirestormBridge', 'AvatarRenderInfo', 'ParcelPropertiesRequest', 'GetObjectCost',
    'RequestMultipleObjects', 'GetObjectPhysicsData', 'GetExperienceInfo', 'RequestTaskInventory', 'AgentRequestSit',
    'MuteListRequest', 'UpdateMuteListEntry', 'RemoveMuteListEntry', 'RequestImage',
    'AgentThrottle', 'UseCircuitCode', 'AgentWearablesRequest', 'AvatarPickerRequest', 'CloseCircuit',
    'CompleteAgentMovement', 'RegionHandshakeReply', 'LogoutRequest', 'ParcelPropertiesRequest',
    'ParcelPropertiesRequestByID', 'MapBlockRequest', 'MapLayerRequest', 'MapItemRequest', 'MapNameRequest',
    'ParcelAccessListRequest', 'AvatarPropertiesRequest', 'DirFindQuery',
    'SetAlwaysRun', 'GetDisplayNames', 'ViewerMetrics', 'AgentResume', 'AgentPause',
    'ViewerAsset', 'GetTexture', 'UUIDNameRequest', 'AgentUpdate', 'AgentAnimation'
    # Would just be confusing for everyone
    'ImprovedInstantMessage',
    # Xfer system isn't authed to begin with, and duping Xfers can lead to premature file deletion. Skip.
    'RequestXfer', 'ConfirmXferPacket', 'AbortXfer', 'SendXferPacket',
}

# Messages that _must_ be sent normally, but are worth mirroring onto the target session to see how
# they would respond
MIRROR = {
    'RequestObjectPropertiesFamily', 'ObjectSelect', 'RequestObjectProperties', 'TransferRequest',
    'RequestMultipleObjects', 'RequestTaskInventory', 'FetchInventory2', 'ScriptDialogReply',
    'ObjectDeselect', 'GenericMessage', 'ChatFromViewer'
}

for msg_name in DEFAULT_TEMPLATE_DICT.message_templates.keys():
    # There are a lot of these.
    if msg_name.startswith("Group") and msg_name.endswith("Request"):
        MIRROR.add(msg_name)


class MessageMirrorAddon(BaseAddon):
    mirror_target_agent: Optional[UUID] = SessionProperty(None)
    mirror_use_target_session: bool = SessionProperty(True)
    mirror_use_target_agent: bool = SessionProperty(True)

    @handle_command(target_agent=Parameter(UUID, optional=True))
    async def mirror_to(self, session: Session, _region, target_agent: Optional[UUID] = None):
        """
        Send this session's outbound messages over another proxied agent's circuit
        """
        if target_agent:
            if target_agent == session.agent_id:
                show_message("Can't mirror our own session")
                target_agent = None
            elif not any(s.agent_id == target_agent for s in session.session_manager.sessions):
                show_message(f"No active proxied session for agent {target_agent}")
                target_agent = None

        self.mirror_target_agent = target_agent
        if target_agent:
            show_message(f"Mirroring to {target_agent}")
        else:
            show_message("Message mirroring disabled")

    @handle_command(enabled=parse_bool)
    async def set_mirror_use_target_session(self, _session, _region, enabled):
        """Replace the original session ID with the target session's ID when mirroring"""
        self.mirror_use_target_session = enabled

    @handle_command(enabled=parse_bool)
    async def set_mirror_use_target_agent(self, _session, _region, enabled):
        """Replace the original agent ID with the target agent's ID when mirroring"""
        self.mirror_use_target_agent = enabled

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.direction != Direction.OUT:
            return

        if not self.mirror_target_agent:
            return

        if message.name in SEND_NORMALLY:
            return

        target_session = None
        for poss_session in session.session_manager.sessions:
            if poss_session.agent_id == self.mirror_target_agent:
                target_session = poss_session

        if not target_session:
            print("Couldn't find target session?")
            return

        target_region = None
        for poss_region in target_session.regions:
            if poss_region.circuit_addr == region.circuit_addr:
                target_region = poss_region

        if not target_region:
            print("Couldn't find equivalent target region?")
            return

        # Send the message normally first if we're mirroring
        if message.name in MIRROR:
            region.circuit.send(message)

        # We're going to send the message on a new circuit, we need to take
        # it so we get a new packet ID and clean ACKs
        message = message.take()

        self._lludp_fixups(target_session, message)
        target_region.circuit.send(message)
        return True

    def _lludp_fixups(self, target_session: Session, message: Message):
        if "AgentData" in message:
            agent_block = message["AgentData"][0]
            if "AgentID" in agent_block and self.mirror_use_target_agent:
                agent_block["AgentID"] = target_session.agent_id
            if "SessionID" in agent_block and self.mirror_use_target_session:
                agent_block["SessionID"] = target_session.id

        if message.name == "TransferRequest":
            transfer_block = message["TransferInfo"][0]
            # This is a duplicated message so we need to give it a new ID
            transfer_block["TransferID"] = UUID.random()
            params = transfer_block.deserialize_var("Params")
            # This kind of Transfer might not even use agent credentials
            if self.mirror_use_target_agent and hasattr(params, 'AgentID'):
                params.AgentID = target_session.agent_id
            if self.mirror_use_target_session and hasattr(params, 'SessionID'):
                params.SessionID = target_session.id
            transfer_block.serialize_var("Params", params)

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        # Already mirrored, ignore.
        if flow.is_replay:
            return

        cap_data = flow.cap_data
        if not cap_data:
            return
        if cap_data.cap_name in SEND_NORMALLY:
            return

        if cap_data.asset_server_cap:
            return
        # Likely doesn't have an exact equivalent in the target session, this is a temporary
        # cap like an uploader URL or a stats URL.
        if cap_data.type == CapType.TEMPORARY:
            return

        session: Optional[Session] = cap_data.session and cap_data.session()
        if not session:
            return

        region: Optional[ProxiedRegion] = cap_data.region and cap_data.region()
        if not region:
            return

        # Session-scoped, so we need to know if we have a session before checking
        if not self.mirror_target_agent:
            return

        target_session: Optional[Session] = None
        for poss_session in session.session_manager.sessions:
            if poss_session.agent_id == self.mirror_target_agent:
                target_session = poss_session
        if not target_session:
            return

        caps_source = target_session
        target_region: Optional[ProxiedRegion] = None
        if region:
            target_region = None
            for poss_region in target_session.regions:
                if poss_region.circuit_addr == region.circuit_addr:
                    target_region = poss_region

            if not target_region:
                print("No region in cap?")
                return
            caps_source = target_region

        new_base_url = caps_source.cap_urls.get(cap_data.cap_name)
        if not new_base_url:
            print("No equiv cap?")
            return

        if cap_data.cap_name in MIRROR:
            flow = flow.copy()

        # Have the cap data reflect the new URL we're pointing at
        flow.metadata["cap_data"] = CapData(
            cap_name=cap_data.cap_name,
            region=weakref.ref(target_region) if target_region else None,
            session=weakref.ref(target_session),
            base_url=new_base_url,
        )

        # Tack any params onto the new base URL for the cap
        new_url = new_base_url + flow.request.url[len(cap_data.base_url):]
        flow.request.url = new_url

        if cap_data.cap_name in MIRROR:
            self._replay_flow(flow, session.session_manager)

    def _replay_flow(self, flow: HippoHTTPFlow, session_manager: SessionManager):
        # Work around mitmproxy bug, changing the URL updates the Host header, which may
        # cause it to drop the port even when it shouldn't have. Fix the host header.
        if flow.request.port not in (80, 443) and ":" not in flow.request.host_header:
            flow.request.host_header = f"{flow.request.host}:{flow.request.port}"
        # Should get repopulated when it goes back through the MITM addon
        flow.metadata.pop("cap_data_ser", None)
        flow.metadata.pop("cap_data", None)
        proxy_queue = session_manager.flow_context.to_proxy_queue
        proxy_queue.put_nowait(("replay", None, flow.get_state()))


addons = [MessageMirrorAddon()]



import os
import unittest

from hippolyzer.lib.base.mesh import LLMeshSerializer, MeshAsset
import hippolyzer.lib.base.serialization as se

BASE_PATH = os.path.dirname(os.path.abspath(__file__))


class TestMesh(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        # Use a rigged cube SLM from the upload process as a test file
        slm_file = os.path.join(BASE_PATH, "test_resources", "testslm.slm")
        with open(slm_file, "rb") as f:
            cls.slm_bytes = f.read()

    def _get_test_mesh_reader(self):
        return se.BufferReader("!", self.slm_bytes)

    def _parse_test_mesh(self):
        reader = self._get_test_mesh_reader()
        mesh: MeshAsset = reader.read(LLMeshSerializer(parse_segment_contents=True))
        return mesh

    def test_basic(self):
        mesh = self._parse_test_mesh()
        self.assertEqual(mesh.header["lowest_lod"]["offset"], 211)
        self.assertSequenceEqual(mesh.segments["skin"]["joint_names"], ["mPelvis"])

    def test_round_trip(self):
        serializer = LLMeshSerializer()
        writer = se.BufferWriter("!")
        writer.write(serializer, self._parse_test_mesh())
        # Write it at least once on this machine in case the gzip impl changes
        first_buf = writer.copy_buffer()
        writer.clear()
        reader = se.BufferReader("!", first_buf)
        writer.write(serializer, reader.read(serializer))
        second_buf = writer.copy_buffer()
        self.assertEqual(first_buf, second_buf)
        # Dates may not round-trip correctly, but length should always be the same
        self.assertEqual(len(first_buf), len(self.slm_bytes))

    def test_serialize_raw_segments(self):
        serializer = LLMeshSerializer(include_raw_segments=True)
        reader = se.BufferReader("!", self.slm_bytes)
        mesh: MeshAsset = reader.read(serializer)
        mesh.raw_segments['high_lod'] += b"foobar"
        expected_high = mesh.raw_segments['high_lod']
        mesh.segments.pop("high_lod")
        writer = se.BufferWriter("!")
        writer.write(serializer, mesh)
        reader = se.BufferReader("!", writer.copy_buffer())
        mesh: MeshAsset = reader.read(serializer)
        self.assertEqual(mesh.raw_segments['high_lod'], expected_high)

    def test_iter_materials(self):
        mesh = self._parse_test_mesh()
        lod_list = list(mesh.iter_lods())
        self.assertEqual(4, len(lod_list))
        self.assertIsInstance(lod_list[0], list)
        mat_list = list(mesh.iter_lod_materials())
        self.assertEqual(4, len(mat_list))
        self.assertIsInstance(mat_list[0], dict)

    def test_make_default_triangle(self):
        tri = MeshAsset.make_triangle()
        self.assertEqual(0.5, tri.segments['high_lod'][0]['Position'][2].X)
        self.assertEqual(1, tri.header['version'])



"""
Example mesh mangler addon, to be used with local mesh addon.

You can edit this live to apply various transforms to local mesh.
If there are no live local mesh instances, the transforms will be
applied to the mesh before upload.

I personally use manglers to strip bounding box materials you need
to add to give a mesh an arbitrary center of rotation / scaling.
"""
from hippolyzer.lib.base.helpers import reorient_coord
from hippolyzer.lib.base.mesh import MeshAsset
from hippolyzer.lib.proxy.addons import AddonManager

import local_mesh
AddonManager.hot_reload(local_mesh, require_addons_loaded=True)


def _reorient_coord_list(coord_list, orientation, min_val: int | float = 0):
    return [reorient_coord(x, orientation, min_val) for x in coord_list]


def reorient_mesh(orientation):
    # Returns a callable that will change `mesh` to match `orientation`
    # X=1, Y=2, Z=3
    def _reorienter(mesh: MeshAsset):
        for material in mesh.iter_lod_materials():
            if "Position" not in material:
                # Must be a NoGeometry LOD
                continue
            # We don't need to use positions_(to/from)_domain here since we're just naively
            # flipping the axes around.
            material["Position"] = _reorient_coord_list(material["Position"], orientation)
            # Are you even supposed to do this to the normals?
            material["Normal"] = _reorient_coord_list(material["Normal"], orientation, min_val=-1)
        return mesh
    return _reorienter


class ExampleMeshManglerAddon(local_mesh.BaseMeshManglerAddon):
    MESH_MANGLERS = [
        # Negate the X and Y axes on any mesh we upload or create temp
        reorient_mesh((-1, -2, 3)),
    ]


addons = [ExampleMeshManglerAddon()]



"""
Drop outgoing packets that might leak what you're looking at, similar to Firestorm
"""

from hippolyzer.lib.base.templates import ViewerEffectType
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


BLOCKED_EFFECTS = (
    ViewerEffectType.EFFECT_LOOKAT,
    ViewerEffectType.EFFECT_BEAM,
    ViewerEffectType.EFFECT_POINTAT,
    ViewerEffectType.EFFECT_EDIT,
)


def handle_lludp_message(_session: Session, region: ProxiedRegion, msg: Message):
    if msg.name == "ViewerEffect" and msg.direction == Direction.OUT:
        new_blocks = [b for b in msg["Effect"] if b["Type"] not in BLOCKED_EFFECTS]
        if new_blocks:
            msg["Effect"] = new_blocks
        else:
            # drop `ViewerEffect` entirely if left with no blocks
            region.circuit.drop_message(msg)
            # Short-circuit any other addons processing this message
            return True



"""
Recapitator addon, merges a base head shape into body shapes.

Only works if both the base shapes and shapes you need to edit are modify.

Useful if you switch heads a lot. Most heads come with a base shape you
have to start from if you don't want the head to look like garbage. If you
have an existing shape for your body, you have to write down all the values
of the base shape's head sliders and edit them onto your body shapes.

This addon does basically the same thing by intercepting shape uploads. After
enabling recapitation, you save the base head shape once. Then the next time you
edit and save a body shape, it will be saved with the head sliders from your base
shape.
"""
import logging
from typing import *

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.templates import AssetType, WearableType
from hippolyzer.lib.base.wearables import Wearable, VISUAL_PARAMS
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty, AssetAliasTracker, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


# Get all VisualParam IDs that belong to head sliders
HEAD_EDIT_GROUPS = ("shape_head", "shape_eyes", "shape_ears", "shape_nose", "shape_mouth", "shape_chin")
HEAD_PARAM_IDS = [v.id for v in VISUAL_PARAMS if v.edit_group in HEAD_EDIT_GROUPS]


class RecapitatorAddon(BaseAddon):
    transaction_remappings: AssetAliasTracker = SessionProperty(AssetAliasTracker)
    recapitating: bool = SessionProperty(bool)
    recapitation_mappings: Dict[int, float] = SessionProperty(dict)

    @handle_command()
    async def enable_recapitation(self, _session: Session, _region: ProxiedRegion):
        """Apply base head shape when saving subsequent shapes"""
        self.recapitating = True
        self.recapitation_mappings.clear()
        show_message("Recapitation enabled, wear the base shape containing the head parameters and save it.")

    @handle_command()
    async def disable_recapitation(self, _session: Session, _region: ProxiedRegion):
        self.recapitating = False
        show_message("Recapitation disabled")

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if not self.recapitating:
            return
        if message.direction != Direction.OUT:
            return
        if message.name != "AssetUploadRequest":
            return
        if message["AssetBlock"]["Type"] != AssetType.BODYPART:
            return

        # Pending asset upload for a bodypart asset. Take the message and request
        # it from the client ourself so we can see what it wants to upload
        new_message = message.take()
        self._schedule_task(self._proxy_bodypart_upload(session, region, new_message))
        return True

    async def _proxy_bodypart_upload(self, session: Session, region: ProxiedRegion, message: Message):
        asset_block = message["AssetBlock"]
        # Asset will already be in the viewer's VFS as the expected asset ID, calculate it.
        asset_id = session.transaction_to_assetid(asset_block["TransactionID"])
        success = False
        try:
            # Xfer the asset from the viewer if it wasn't small enough to fit in AssetData
            if asset_block["AssetData"]:
                asset_data = asset_block["AssetData"]
            else:
                xfer = await region.xfer_manager.request(
                    vfile_id=asset_id,
                    vfile_type=AssetType.BODYPART,
                    direction=Direction.IN,
                )
                asset_data = xfer.reassemble_chunks()

            wearable = Wearable.from_bytes(asset_data)
            # If they're uploading a shape, process it.
            if wearable.wearable_type == WearableType.SHAPE:
                if self.recapitation_mappings:
                    # Copy our previously saved head params over
                    for key, value in self.recapitation_mappings.items():
                        wearable.parameters[key] = value
                    # Upload the changed version
                    asset_data = wearable.to_bytes()
                    show_message("Recapitated shape")
                else:
                    # Don't have a recapitation mapping yet, use this shape as the base.
                    for param_id in HEAD_PARAM_IDS:
                        self.recapitation_mappings[param_id] = wearable.parameters[param_id]
                    show_message("Got base parameters for recapitation, head parameters will be copied")

            # Upload it ourselves with a new transaction ID that can be traced back to
            # the original. This is important because otherwise the viewer will use its
            # own cached version of the shape, under the assumption it wasn't modified
            # during upload.
            new_transaction_id = self.transaction_remappings.get_alias_uuid(
                asset_block["TransactionID"]
            )
            await region.xfer_manager.upload_asset(
                asset_type=AssetType.BODYPART,
                data=asset_data,
                transaction_id=new_transaction_id,
            )
            success = True
        except:
            logging.exception("Exception while recapitating")
        # Tell the viewer about the status of its original upload
        region.circuit.send(Message(
            "AssetUploadComplete",
            Block("AssetBlock", UUID=asset_id, Type=asset_block["Type"], Success=success),
            direction=Direction.IN,
        ))

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        # Skip requests that aren't related to patching an existing item
        if flow.cap_data.cap_name != "InventoryAPIv3":
            return
        if flow.request.method != "PATCH":
            return
        if "/item/" not in flow.request.url:
            return

        parsed = llsd.parse_xml(flow.request.content)
        if parsed.get("type") != "bodypart":
            return
        # `hash_id` being present means we're updating the item to point to a newly
        # uploaded asset. It's actually a transaction ID.
        transaction_id: Optional[UUID] = parsed.get("hash_id")
        if not transaction_id:
            return
        # We have an original transaction ID, do we need to remap it to an alias ID?
        orig_id = self.transaction_remappings.get_alias_uuid(transaction_id, create=False)
        if not orig_id:
            return

        parsed["hash_id"] = orig_id
        flow.request.content = llsd.format_xml(parsed)


addons = [RecapitatorAddon()]




"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import unittest

from hippolyzer.lib.base.message.llsd_msg_serializer import LLSDMessageSerializer
from hippolyzer.lib.base.message.message import Message, Block


class TestLLSDSerializer(unittest.TestCase):
    def setUp(self):
        self.serializer = LLSDMessageSerializer()
        self.test_msg = Message(
            "EnableSimulator",
            Block(
                "SimulatorInfo",
                Handle=200,
                IP="127.0.0.1",
                Port=201,
            ),
        )
        self.test_dict = {
            "message": "EnableSimulator",
            "body": {
                "SimulatorInfo": [{
                    "Handle": b'\x00\x00\x00\x00\x00\x00\x00\xc8',
                    "IP": b"\x7f\x00\x00\x01",
                    "Port": 201,
                }]
            }
        }
        self.test_xml = b'<?xml version="1.0" ?><llsd><map><key>message</key><string>EnableSimulator</string>' \
                        b'<key>body</key><map><key>SimulatorInfo</key><array><map><key>Handle</key><binary>' \
                        b'AAAAAAAAAMg=</binary><key>IP</key><binary>fwAAAQ==</binary><key>Port</key><integer>' \
                        b'201</integer></map></array></map></map></llsd>'

    def test_dict_serialization(self):
        msg_dict = self.serializer.serialize(self.test_msg, as_dict=True)
        self.assertEqual(msg_dict, self.test_dict)

    def test_xml_serialization(self):
        msg_xml = self.serializer.serialize(self.test_msg)
        self.assertEqual(msg_xml, self.test_xml)

    def test_dict_deserialization(self):
        msg = self.serializer.deserialize(self.test_dict)
        self.assertEqual(self.test_msg.to_dict(), msg.to_dict())

    def test_xml_deserialization(self):
        msg = self.serializer.deserialize(self.test_xml)
        self.assertEqual(self.test_msg.to_dict(), msg.to_dict())



"""
Debugger for detecting when animations within an object get started or stopped

Useful for tracking down animation sequence-related bugs within your LSL scripts,
or debugging automatic animation stopping behavior in the viewer.

If an animation unexpectedly stops and nobody requested it be stopped, it's a potential viewer bug (or priority issue).
If an animation unexpectedly stops and the viewer requested it be stopped, it's also a potential viewer bug.
If an animation unexpectedly stops and only the server requested it be stopped, it's a potential script / server bug.
"""

from typing import *

from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.base.templates import AssetType
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.addon_utils import show_message


class AnimTrackerAddon(BaseAddon):
    should_track_anims: bool = SessionProperty(False)
    anims_lookup: Dict[UUID, str] = SessionProperty(dict)
    last_tracker_anims: Set[UUID] = SessionProperty(set)

    def _format_anim_diffs(self, started_anims: Set[UUID], stopped_anims: Set[UUID]):
        added_strs = [f"+{self.anims_lookup[x]!r}" for x in started_anims]
        removed_strs = [f"-{self.anims_lookup[x]!r}" for x in stopped_anims]

        return ", ".join(removed_strs + added_strs)

    @handle_command()
    async def track_anims(self, session: Session, region: ProxiedRegion):
        """Track when animations within this object get started or stopped"""
        if self.should_track_anims:
            self.last_tracker_anims.clear()
            self.anims_lookup.clear()

        selected = region.objects.lookup_localid(session.selected.object_local)
        if not selected:
            return

        self.should_track_anims = True

        object_items = await region.objects.request_object_inv(selected)

        anims: Dict[UUID, str] = {}
        for item in object_items:
            if item.type != AssetType.ANIMATION:
                continue
            anims[item.true_asset_id] = item.name

        self.anims_lookup = anims

    @handle_command()
    async def stop_tracking_anims(self, _session: Session, _region: ProxiedRegion):
        """Stop reporting differences"""
        if self.should_track_anims:
            self.should_track_anims = False
            self.last_tracker_anims.clear()
            self.anims_lookup.clear()

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if not self.should_track_anims:
            return

        if message.name != "AgentAnimation" or message.direction != Direction.OUT:
            # AgentAnimation is the message the viewer uses to request manually starting or stopping animations.
            # We don't care about other messages, we're just interested in distinguishing cases where the viewer
            # specifically requested something vs something being done by the server on its own.
            return
        av = session.objects.lookup_avatar(session.agent_id)
        if not av or not av.Object:
            print("Somehow didn't know about our own av object?")
            return

        current_anims = set([x for x in av.Object.Animations if x in self.anims_lookup])
        started_anims: Set[UUID] = set()
        stopped_anims: Set[UUID] = set()

        for block in message["AnimationList"]:
            anim_id = block["AnimID"]
            if anim_id not in self.anims_lookup:
                continue

            start_anim = block["StartAnim"]
            already_started = anim_id in current_anims
            if start_anim == already_started:
                # No change
                continue

            if start_anim:
                started_anims.add(anim_id)
            else:
                stopped_anims.add(anim_id)

        if started_anims or stopped_anims:
            show_message("Viewer Requested Anims: " + self._format_anim_diffs(started_anims, stopped_anims))

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        if not self.should_track_anims:
            return
        if obj.FullID != session.agent_id:
            return
        if "Animations" not in updated_props:
            return

        current_anims = set([x for x in obj.Animations if x in self.anims_lookup])
        started_anims = current_anims - self.last_tracker_anims
        stopped_anims = self.last_tracker_anims - current_anims

        self.last_tracker_anims.clear()
        self.last_tracker_anims.update(current_anims)

        if started_anims or stopped_anims:
            show_message("Anim Diffs: " + self._format_anim_diffs(started_anims, stopped_anims))


addons = [AnimTrackerAddon()]



"""
Try and diagnose very slow avatar appearance loads when the avatars first come on the scene

I guess use LEAP or something to detect when things _actually_ declouded.
"""
from typing import *

import dataclasses
import datetime as dt

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.base.templates import PCode
from hippolyzer.lib.proxy.addon_utils import BaseAddon, GlobalProperty
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


@dataclasses.dataclass
class AvatarBakeRequest:
    requested: dt.datetime
    received: Optional[dt.datetime] = None


@dataclasses.dataclass
class AvatarAppearanceRecord:
    object_received: dt.datetime
    """When we learned about the agent as an object"""
    appearance_received: Optional[dt.datetime] = None
    """When AvatarAppearance was first received"""
    bake_requests: Dict[str, AvatarBakeRequest] = dataclasses.field(default_factory=dict)
    """Layer name -> request / response details"""


class AppearanceDelayTrackerAddon(BaseAddon):
    # Should be able to access this in the REPL
    # Normally we'd use a session property, but we may not have a proper session context for some requests
    av_appearance_data: Dict[UUID, AvatarAppearanceRecord] = GlobalProperty(dict)

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        if obj.PCode == PCode.AVATAR and obj.FullID not in self.av_appearance_data:
            self.av_appearance_data[obj.FullID] = AvatarAppearanceRecord(object_received=dt.datetime.now())

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.name != "AvatarAppearance":
            return
        agent_id = message["Sender"]["ID"]
        appearance_data = self.av_appearance_data.get(agent_id)
        if not appearance_data:
            print(f"Got appearance for {agent_id} without knowing about object?")
            return

        if appearance_data.appearance_received:
            return
        appearance_data.appearance_received = dt.datetime.now()

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if not flow.cap_data:
            return
        if flow.cap_data.cap_name != "AppearanceService":
            return

        agent_id = UUID(flow.request.url.split('/')[-3])
        slot_name = flow.request.url.split('/')[-2]
        appearance_data = self.av_appearance_data.get(agent_id)
        if not appearance_data:
            print(f"Got AppearanceService req for {agent_id} without knowing about object?")
            return
        if slot_name in appearance_data.bake_requests:
            # We already requested this slot before
            return
        appearance_data.bake_requests[slot_name] = AvatarBakeRequest(requested=dt.datetime.now())

    def handle_http_response(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if not flow.cap_data:
            return
        if flow.cap_data.cap_name != "AppearanceService":
            return

        agent_id = UUID(flow.request.url.split('/')[-3])
        slot_name = flow.request.url.split('/')[-2]
        appearance_data = self.av_appearance_data.get(agent_id)
        if not appearance_data:
            return
        slot_details = appearance_data.bake_requests.get(slot_name)
        if not slot_details:
            return
        slot_details.received = dt.datetime.now()


addons = [AppearanceDelayTrackerAddon()]



from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class REPLExampleAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.name == "ChatFromViewer":
            chat_msg = message["ChatData"]["Message"]
            if not chat_msg:
                return
            # Intercept chat messages containing "hippolyzer_test" as an example
            if "hippolyzer_test" in chat_msg:
                if AddonManager.have_active_repl():
                    # Already intercepting, don't touch it
                    return
                # Take ownership of the message so it won't be sent by the
                # usual machinery.
                _new_msg = message.take()
                # repl will have access to `_new_msg` and can send it with
                # `region.circuit.send_message()` after it's modified.
                AddonManager.spawn_repl()
                return True
            if "hippolyzer_async_test" in chat_msg:
                if AddonManager.have_active_repl():
                    # Already intercepting, don't touch it
                    return

                async def _coro():
                    foo = 4
                    # spawn_repl() can be `await`ed, changing foo
                    # in the repl will change what's printed on exit.
                    await AddonManager.spawn_repl()
                    print("foo is", foo)

                self._schedule_task(_coro())


addons = [REPLExampleAddon()]



"""
Test client handling of messages with malformed bodies

Serializes message, but mutates parts of the message body before the ACKs.

You don't want to use this unless you're a viewer developer trying to fix bugs.
There's a 95% chance it will crash your viewer, and maybe make you teleport random
places. Definitely don't test it while logged in with an account with access to
anything important.
"""
import copy
import datetime as dt
import logging
import random

from hippolyzer.lib.base.message.msgtypes import PacketLayout
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session

LOG = logging.getLogger(__name__)


class PacketMutationAddon(BaseAddon):
    def __init__(self):
        self.serializer = UDPMessageSerializer()

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        # Only inbound messages, don't fiddle with the sim.
        if message.direction != Direction.IN:
            return
        # Messing with these may kill your circuit
        if message.name in ("PacketAck", "StartPingCheck", "CompletePingCheck"):
            return

        # Give login time to complete
        if session.started_at + dt.timedelta(seconds=10) > dt.datetime.now():
            return

        # Do it randomly
        if random.random() < 0.5:
            return

        # We need to take this message because we're going to
        # send our own re-serialized version. Don't use take_message()
        # because we're going to keep packet_id and acks the same.
        prepared = copy.deepcopy(message)
        # This message only had ACKs that were dropped.
        if not region.circuit.prepare_message(prepared):
            return

        serialized = bytearray(self.serializer.serialize(prepared))

        # Figure out where the ACKs will be so we don't mess those up
        acks_size = 0
        if prepared.acks:
            acks_size = 1 + (len(prepared.acks) * 4)

        # Give enough space for the message name, and ignore the acks at the end
        # mesage name can be 5 bytes if zerocoded.
        body_slice = slice(PacketLayout.PHL_NAME + 5 + message.offset, -1 - acks_size)
        body_view = serialized[body_slice]
        # The message is too small and we're left with nothing.
        if not body_view:
            return

        # Can be switched out with _flip_body_bytes() or something
        changed_body = self._truncate_body(body_view)
        if changed_body is None:
            return
        serialized[body_slice] = changed_body

        # Send out the raw mutated datagram
        region.circuit.send_datagram(serialized, message.direction)
        # Tell the proxy that we already sent the message and to short-circuit
        return True

    def _truncate_body(self, body_view: bytearray):
        # Don't want to mess with bodies this short.
        if len(body_view) < 4:
            return
        # Slice off the last bit of the body
        del body_view[int(len(body_view) * 0.7):-1]
        return body_view

    def _flip_body_bytes(self, body_view: bytearray):
        # Don't want to mess with bodies this short.
        if len(body_view) < 4:
            return

        # randomly flip bytes up to 19 bytes away from the end
        for i in range(-19, 0):
            if random.random() < 0.3:
                body_view[i] = (~body_view[i]) & 0xFF
        return body_view


addons = [PacketMutationAddon()]



import unittest

from hippolyzer.lib.proxy.commands import handle_command, Parameter


class ExampleCommandHandler:
    def __init__(self):
        self.bar = None
        self.baz = None

    @handle_command("foobar", bar=str, baz=int)
    async def foo(self, _session, _region, bar, baz):
        self.bar = bar
        self.baz = baz

    @handle_command(
        "bazquux",
        x=Parameter(sep="/", parser=float),
        y=Parameter(sep="/", parser=float),
        z=Parameter(parser=float),
    )
    async def quux(self, _session, _region, x, y, z):
        self.bar = (x, y, z)

    @handle_command(
        y=str,
    )
    async def own_name(self, _session, _region, y):
        pass

    @handle_command(
        x=Parameter(str, optional=True),
    )
    async def optional(self, _session, _region, x=42):
        self.bar = x


class TestCommandHandlers(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.handler = ExampleCommandHandler()

    async def test_foo(self):
        await self.handler.foo(None, None, "foozy 1")
        self.assertEqual(self.handler.bar, "foozy")
        self.assertEqual(self.handler.baz, 1)

    async def test_quux(self):
        # Alternative command name was provided
        self.assertEqual(self.handler.quux.command.name, "bazquux")
        await self.handler.quux(None, None, "1/2/3")
        self.assertEqual(self.handler.bar, (1, 2, 3))

    async def test_own_name(self):
        self.assertEqual(self.handler.own_name.command.name, "own_name")

    async def test_missing_param(self):
        with self.assertRaises(KeyError):
            await self.handler.foo(None, None, "")

    async def test_optional_param(self):
        await self.handler.optional(None, None, "foo")  # type: ignore
        self.assertEqual(self.handler.bar, "foo")
        await self.handler.optional(None, None, "")  # type: ignore
        # Should have picked up the default value
        self.assertEqual(self.handler.bar, 42)

    async def test_bad_command(self):
        with self.assertRaises(ValueError):
            class _BadCommandHandler:
                @handle_command("foobaz")
                def bad_command(self, session, region):
                    assert False



"""
Allows specifying a target object to apply a mesh preview to. When a local mesh target
is specified, hitting the "calculate upload cost" button in the mesh uploader will instead
apply the mesh to the local mesh target. It works on attachments too. Useful for testing rigs before a
final, real upload.

Select an object and do /524 set_local_mesh_target, then go through the mesh upload flow.
Mesh pieces will be mapped to your object based on object name. Note that if you're using Blender
these will be based on the name of your _geometry nodes_ and not the objects themselves.

The object you select as a mesh target must contain a mesh prim. The mesh objects you use as a local
mesh target should have at least as many faces as the mesh you want to apply to it or you won't
be able to set textures on those faces correctly.

When you're done with local mesh and want to allow regular mesh upload again, do
/524 disable_local_mesh

Does not attempt to apply textures uploaded with the mesh.
"""
from __future__ import annotations

import ctypes
import secrets
from typing import *

import mitmproxy.http

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.mesh import LLMeshSerializer, MeshAsset
from hippolyzer.lib.base import serialization as se
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.base.templates import ExtraParamType
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.addon_utils import show_message, BaseAddon, GlobalProperty, SessionProperty
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.http_asset_repo import HTTPAssetRepo
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


def _modify_crc(crc_tweak, crc_val):
    return ctypes.c_uint32(crc_val ^ crc_tweak).value


def _mangle_mesh_list(mesh_list: List[bytes], manglers: List[Callable]) -> List[bytes]:
    if not mesh_list or not manglers:
        return mesh_list
    new_mesh_list = []
    mesh_ser = LLMeshSerializer()
    for mesh_bytes in mesh_list:
        reader = se.BufferReader("!", mesh_bytes)
        mesh: MeshAsset = reader.read(mesh_ser)
        for mangler in manglers:
            mesh = mangler(mesh)
        writer = se.BufferWriter("!")
        writer.write(mesh_ser, mesh)
        new_mesh_list.append(writer.copy_buffer())
    return new_mesh_list


class MeshUploadInterceptingAddon(BaseAddon):
    mesh_crc_tweak: int = GlobalProperty(default=secrets.randbits(32))
    mesh_manglers: List[Callable[[MeshAsset], MeshAsset]] = GlobalProperty(list)
    # LocalIDs being targeted for local mesh
    local_mesh_target_locals: List[int] = SessionProperty(list)
    # Name -> mesh index mapping
    local_mesh_mapping: Dict[str, int] = SessionProperty(dict)
    # originally supplied mesh bytes, indexed by mesh index.
    # mostly used for re-mangling if mesh manglers changed.
    local_mesh_orig_bytes: List[bytes] = SessionProperty(list)
    # Above, but for the local asset IDs
    local_mesh_asset_ids: List[UUID] = SessionProperty(list)

    def handle_init(self, session_manager: SessionManager):
        # Other plugins can add to this list to apply transforms to mesh
        # whenever it's uploaded.
        self.remangle_local_mesh(session_manager)

    @handle_command()
    async def set_local_mesh_target(self, session: Session, region: ProxiedRegion):
        """Set the currently selected objects as the target for local mesh"""
        selected_links = [region.objects.lookup_localid(l_id) for l_id in session.selected.object_locals]
        selected_links = [o for o in selected_links if o is not None]
        if not selected_links:
            show_message("Nothing selected")
            return
        old_locals = self.local_mesh_target_locals
        self.local_mesh_target_locals = [
            x.LocalID
            for x in selected_links
            if ExtraParamType.MESH in x.ExtraParams
        ]

        if old_locals:
            # Return the old objects to normal
            self.mesh_crc_tweak = secrets.randbits(32)
            region.objects.request_objects(old_locals)

        if not self.local_mesh_target_locals:
            show_message("There must be at least one mesh object in the linkset!")
            return

        # We'll need the name for all of these to pick which mesh asset to
        # apply to them.
        region.objects.request_object_properties(self.local_mesh_target_locals)
        show_message(f"Targeting {self.local_mesh_target_locals}")

    @handle_command()
    async def disable_local_mesh(self, session: Session, region: ProxiedRegion):
        """Disable local mesh mode, allowing mesh upload and returning targets to normal"""
        # Put the target objects back to normal and kill the temp assets
        old_locals = tuple(self.local_mesh_target_locals)
        self.local_mesh_target_locals.clear()
        asset_repo: HTTPAssetRepo = session.session_manager.asset_repo
        for asset_id in self.local_mesh_asset_ids:
            del asset_repo[asset_id]
        self.local_mesh_asset_ids.clear()
        self.local_mesh_asset_ids.clear()
        self.local_mesh_mapping.clear()
        if old_locals:
            region.objects.request_objects(old_locals)
        show_message(f"Cleared target {old_locals}")

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        # Replace any mesh asset IDs in tracked objects with our local assets
        if not self.local_mesh_target_locals:
            return
        if not self.local_mesh_asset_ids:
            return

        if message.name == "ObjectUpdate":
            for block in message["ObjectData"]:
                if block["ID"] not in self.local_mesh_target_locals:
                    continue
                block["CRC"] = _modify_crc(self.mesh_crc_tweak, block["CRC"])
                parsed_params = block.deserialize_var("ExtraParams")
                if not parsed_params:
                    continue
                obj = region.objects.lookup_localid(block["ID"])
                if not obj:
                    return
                parsed_params[ExtraParamType.MESH]["Asset"] = self._pick_mesh_asset(obj)
                block.serialize_var("ExtraParams", parsed_params)
        elif message.name == "ObjectUpdateCompressed":
            for block in message["ObjectData"]:
                update_data = block.deserialize_var("Data")
                if not update_data:
                    continue
                if update_data["ID"] not in self.local_mesh_target_locals:
                    continue
                update_data["CRC"] = _modify_crc(self.mesh_crc_tweak, update_data["CRC"])
                if not update_data.get("ExtraParams"):
                    continue

                obj = region.objects.lookup_localid(update_data["ID"])
                if not obj:
                    return
                extra_params = update_data["ExtraParams"]
                extra_params[ExtraParamType.MESH]["Asset"] = self._pick_mesh_asset(obj)
                block.serialize_var("Data", update_data)

    def _pick_mesh_asset(self, obj: Object) -> UUID:
        mesh_idx = self.local_mesh_mapping.get(obj.Name, 0)
        # Use whatever the first mesh was if we don't have a match on name.
        return self.local_mesh_asset_ids[mesh_idx]

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        cap_data = flow.cap_data
        if not cap_data:
            return
        if cap_data.cap_name == "NewFileAgentInventory":
            # Might be an upload cost calculation request for mesh, includes the actual mesh data.
            payload = llsd.parse_xml(flow.request.content)
            if "asset_resources" not in payload:
                return

            orig_mesh_list = payload["asset_resources"].get("mesh_list")
            if not orig_mesh_list:
                return

            # Replace the mesh instances in the payload with versions run through our mangler
            new_mesh_list = _mangle_mesh_list(orig_mesh_list, self.mesh_manglers)
            payload["asset_resources"]["mesh_list"] = new_mesh_list

            # We have local mesh instances, re-use the data sent along with the upload cost
            # request to apply the mesh to our local mesh objects intead.
            if self.local_mesh_target_locals:
                region: ProxiedRegion = cap_data.region()
                asset_repo: HTTPAssetRepo = session_manager.asset_repo
                # Apply the new mesh to any local mesh targets
                self._replace_local_mesh(region, asset_repo, new_mesh_list)
                # Keep the original bytes around in case manglers get reloaded
                # and we want to re-run them
                self.local_mesh_orig_bytes = orig_mesh_list
                instances = payload["asset_resources"]["instance_list"]
                # To figure out what mesh index applies to what object name
                self.local_mesh_mapping = {x["mesh_name"]: x["mesh"] for x in instances}

                # Fake a response, we don't want to actually send off the request.
                flow.response = mitmproxy.http.Response.make(
                    200,
                    b"",
                    {
                        "Content-Type": "text/plain",
                        "Connection": "close",
                    }
                )
                show_message("Applying local mesh")
            # Even if we're not in local mesh mode, we want the upload cost for
            # our mangled mesh
            elif self.mesh_manglers:
                flow.request.content = llsd.format_xml(payload)
                show_message("Mangled upload cost request")
        elif cap_data.cap_name == "NewFileAgentInventoryUploader":
            # Don't bother looking at this if we have no manglers
            if not self.mesh_manglers:
                return
            # Depending on what asset is being uploaded the body may not even be LLSD.
            if not flow.request.content or b"mesh_list" not in flow.request.content:
                return
            payload = llsd.parse_xml(flow.request.content)
            if not payload.get("mesh_list"):
                return

            payload["mesh_list"] = _mangle_mesh_list(payload["mesh_list"], self.mesh_manglers)
            flow.request.content = llsd.format_xml(payload)
            show_message("Mangled upload request")

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        if obj.LocalID not in self.local_mesh_target_locals:
            return
        if "Name" not in updated_props or obj.Name is None:
            return
        # A local mesh target has a new name, which mesh we need to apply
        # to the object may have changed.
        self.mesh_crc_tweak = secrets.randbits(32)
        region.objects.request_objects(obj.LocalID)

    @classmethod
    def _replace_local_mesh(cls, region: ProxiedRegion, asset_repo, mesh_list: List[bytes]) -> None:
        cls.mesh_crc_tweak = secrets.randbits(32)

        for asset_id in cls.local_mesh_asset_ids:
            del asset_repo[asset_id]
        cls.local_mesh_asset_ids.clear()
        for mesh_blob in mesh_list:
            cls.local_mesh_asset_ids.append(asset_repo.create_asset(mesh_blob))
        # Ask for a full update so we can clobber the mesh param
        # Janky hack around the fact that we don't know how to build
        # them from scratch yet.
        region.objects.request_objects(cls.local_mesh_target_locals)

    @classmethod
    def remangle_local_mesh(cls, session_manager: SessionManager):
        # We want CRCs that are stable for the duration of the session, but will
        # cause a cache miss for objects cached before this session. Generate a
        # random value to XOR all CRCs with
        # We need to regen this when we force a re-mangle to indicate that the
        # viewer should pay attention to the incoming ObjectUpdate
        cls.mesh_crc_tweak = secrets.randbits(32)

        asset_repo: HTTPAssetRepo = session_manager.asset_repo
        # Mesh manglers are global, so we need to re-mangle mesh for all sessions
        for session in session_manager.sessions:
            # Push the context of this session onto the stack so we can access
            # session-scoped properties
            with addon_ctx.push(new_session=session, new_region=session.main_region):
                if not cls.local_mesh_target_locals:
                    continue
                orig_bytes = cls.local_mesh_orig_bytes
                if not orig_bytes:
                    continue
                show_message("Remangling mesh", session=session)
                mesh_list = _mangle_mesh_list(orig_bytes, cls.mesh_manglers)
                cls._replace_local_mesh(session.main_region, asset_repo, mesh_list)


class BaseMeshManglerAddon(BaseAddon):
    """Base class for addons that mangle uploaded or local mesh"""
    MESH_MANGLERS: List[Callable[[MeshAsset], MeshAsset]]

    def handle_init(self, session_manager: SessionManager):
        # Add our manglers into the list
        MeshUploadInterceptingAddon.mesh_manglers.extend(self.MESH_MANGLERS)
        # Tell the local mesh plugin that the mangler list changed, and to re-apply
        MeshUploadInterceptingAddon.remangle_local_mesh(session_manager)

    def handle_unload(self, session_manager: SessionManager):
        # Clean up our manglers before we go away
        mangler_list = MeshUploadInterceptingAddon.mesh_manglers
        for mangler in self.MESH_MANGLERS:
            if mangler in mangler_list:
                mangler_list.remove(mangler)
        MeshUploadInterceptingAddon.remangle_local_mesh(session_manager)


addons = [MeshUploadInterceptingAddon()]



"""
You don't need RLV, we have RLV at home.

RLV at home:
"""

from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.templates import ChatType
from hippolyzer.lib.proxy.addon_utils import BaseAddon, send_chat
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


def send_rlv_chat(channel: int, message: str):
    # We always shout.
    send_chat(channel=channel, message=message, chat_type=ChatType.NORMAL)


class RLVAtHomeAddon(BaseAddon):
    """
    Addon for pretending to be an RLV-enabled viewer

    Useful if you want only a specific subset of RLV and don't want everything RLV normally allows,
    or want to override some RLV builtins.
    """
    def handle_rlv_command(self, session: Session, region: ProxiedRegion, source: UUID,
                           behaviour: str, options: List[str], param: str) -> bool | None:
        # print(behaviour, options, param)
        if behaviour == "clear":
            return True
        elif behaviour in ("versionnum", "versionnew", "version"):
            # People tend to just check that this returned anything at all. Just say we're 2.0.0 for all of these.
            send_rlv_chat(int(param), "2.0.0")
            return True
        elif behaviour == "getinv":
            # Pretend we don't have anything
            send_rlv_chat(int(param), "")
            return True
        elif behaviour == "sit":
            # Sure, we can sit on stuff, whatever.
            region.circuit.send(Message(
                'AgentRequestSit',
                Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
                Block('TargetObject', TargetID=UUID(options[0]), Offset=(0, 0, 0)),
            ))
            return True
        return None


addons = [RLVAtHomeAddon()]



from typing import NamedTuple, List, Sequence

from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.templates import ChatType


class RLVCommand(NamedTuple):
    behaviour: str
    param: str
    options: List[str]


class RLVParser:
    @staticmethod
    def is_rlv_message(msg: Message) -> bool:
        chat: str = msg["ChatData"]["Message"]
        chat_type: int = msg["ChatData"]["ChatType"]
        return chat and chat.startswith("@") and chat_type == ChatType.OWNER

    @staticmethod
    def parse_chat(chat: str) -> List[RLVCommand]:
        assert chat.startswith("@")
        chat = chat.lstrip("@")
        commands = []
        for command_str in chat.split(","):
            if not command_str:
                continue
            # RLV-style command, `<cmd>(:<option1>;<option2>)?(=<param>)?`
            # Roughly (?<behaviour>[^:=]+)(:(?<option>[^=]*))?=(?<param>\w+)
            options, _, param = command_str.partition("=")
            behaviour, _, options = options.partition(":")
            # TODO: Not always correct, commands can specify their own parsing for the option field
            #  maybe special-case these?
            options = options.split(";") if options else []
            commands.append(RLVCommand(behaviour, param, options))
        return commands

    @staticmethod
    def format_chat(commands: Sequence[RLVCommand]) -> str:
        assert commands
        chat = ""
        for command in commands:
            if chat:
                chat += ","

            chat += command.behaviour
            if command.options:
                chat += ":" + ";".join(command.options)
            if command.param:
                chat += "=" + command.param
        return "@" + chat



from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.proxy.addon_utils import send_chat, BaseAddon, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class GreetingAddon(BaseAddon):
    @handle_command()
    async def greetings(self, session: Session, region: ProxiedRegion):
        """Greet everyone around you"""
        our_avatar = region.objects.lookup_avatar(session.agent_id)
        if not our_avatar:
            show_message("Don't have an agent object?")

        # Look this up in the session object store since we may be next
        # to a region border.
        other_avatars = [o for o in session.objects.all_avatars if o.FullID != our_avatar.FullID]

        if not other_avatars:
            show_message("No other avatars?")

        for other_avatar in other_avatars:
            dist = Vector3.dist(our_avatar.GlobalPosition, other_avatar.GlobalPosition)
            if dist >= 19.0:
                continue
            if other_avatar.PreferredName is None:
                continue
            send_chat(f"Greetings, {other_avatar.PreferredName}!")


addons = [GreetingAddon()]



import asyncio
import unittest
from unittest.mock import MagicMock

from hippolyzer.lib.base.events import Event


class TestEvents(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.event = Event()

    async def test_trigger_sync(self):
        mock = MagicMock(return_value=False)
        self.event.subscribe(mock)
        self.event.notify("foo")
        mock.assert_called_with("foo")
        self.assertIn(mock, [x[0] for x in self.event.subscribers])

    async def test_trigger_sync_unsub(self):
        mock = MagicMock(return_value=True)
        self.event.subscribe(mock)
        self.event.notify("foo")
        mock.assert_called_with("foo")
        self.assertNotIn(mock, [x[0] for x in self.event.subscribers])

    async def test_trigger_async(self):
        called = asyncio.Event()
        mock = MagicMock()

        async def _mock_wrapper(*args, **kwargs):
            called.set()
            mock(*args, **kwargs)
        self.event.subscribe(_mock_wrapper)
        self.event.notify("foo")
        await called.wait()
        mock.assert_called_with("foo")
        self.assertIn(_mock_wrapper, [x[0] for x in self.event.subscribers])

    async def test_trigger_async_unsub(self):
        called = asyncio.Event()
        mock = MagicMock()

        async def _mock_wrapper(*args, **kwargs):
            called.set()
            mock(*args, **kwargs)
            return True
        self.event.subscribe(_mock_wrapper)
        self.event.notify("foo")
        await called.wait()
        mock.assert_called_with("foo")
        self.assertNotIn(_mock_wrapper, [x[0] for x in self.event.subscribers])

    async def test_multiple_subscribers(self):
        called = asyncio.Event()
        called2 = asyncio.Event()

        self.event.subscribe(lambda *args: called.set())
        self.event.subscribe(lambda *args: called2.set())

        self.event.notify(None)

        self.assertTrue(called.is_set())
        self.assertTrue(called2.is_set())



"""
Parse the horrible legacy inventory-related format.

It's typically only used for object contents now.
"""

# TODO: Maybe handle CRC calculation? Does anything care about that?
#  I don't think anything in the viewer actually looks at the result
#  of the CRC check for UDP stuff.

from __future__ import annotations

import abc
import asyncio
import dataclasses
import datetime as dt
import inspect
import logging
import secrets
import struct
import weakref
from io import StringIO
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.legacy_schema import (
    parse_schema_line,
    SchemaBase,
    SchemaDate,
    SchemaFieldSerializer,
    SchemaHexInt,
    SchemaInt,
    SchemaLLSD,
    SchemaMultilineStr,
    SchemaParsingError,
    SchemaStr,
    SchemaUUID,
    schema_field,
)
from hippolyzer.lib.base.message.message import Block
from hippolyzer.lib.base.templates import SaleType, InventoryType, LookupIntEnum, AssetType, FolderType

MAGIC_ID = UUID("3c115e51-04f4-523c-9fa6-98aff1034730")
LOG = logging.getLogger(__name__)
_T = TypeVar("_T")


class SchemaFlagField(SchemaHexInt):
    """Like a hex int, but must be serialized as bytes in LLSD due to being a U32"""
    @classmethod
    def from_llsd(cls, val: Any, flavor: str) -> int:
        # Sometimes values in S32 range will just come through normally
        if isinstance(val, int):
            return val

        if flavor == "legacy":
            return struct.unpack("!I", val)[0]
        return val

    @classmethod
    def to_llsd(cls, val: int, flavor: str) -> Any:
        if flavor == "legacy":
            return struct.pack("!I", val)
        return val


class SchemaEnumField(SchemaStr, Generic[_T]):
    def __init__(self, enum_cls: Type[LookupIntEnum]):
        super().__init__()
        self._enum_cls = enum_cls

    def deserialize(self, val: str) -> _T:
        return self._enum_cls.from_lookup_name(val)

    def serialize(self, val: _T) -> str:
        return self._enum_cls(val).to_lookup_name()

    def from_llsd(self, val: Union[str, int], flavor: str) -> _T:
        if flavor == "legacy":
            return self.deserialize(val)
        return self._enum_cls(val)

    def to_llsd(self, val: _T, flavor: str) -> Union[int, str]:
        if flavor == "legacy":
            return self.serialize(val)
        return int(val)


def _yield_schema_tokens(reader: StringIO):
    in_bracket = False
    # empty str == EOF in Python
    while line := reader.readline():
        line = line.strip()
        # Whitespace-only lines are automatically skipped
        if not line:
            continue
        try:
            key, val = parse_schema_line(line)
        except SchemaParsingError:
            # Can happen if there's a malformed multi-line string, just
            # skip by it.
            LOG.warning(f"Found invalid inventory line {line!r}")
            continue
        if key == "{":
            if in_bracket:
                LOG.warning("Found multiple opening brackets inside structure, "
                            "was a nested structure not handled?")
            in_bracket = True
            continue
        if key == "}":
            if not in_bracket:
                LOG.warning("Unexpected closing bracket")
            in_bracket = False
            break
        yield key, val
    if in_bracket:
        LOG.warning("Reached EOF while inside a bracket")


class InventoryBase(SchemaBase):
    SCHEMA_NAME: ClassVar[str]

    @classmethod
    def from_reader(cls, reader: StringIO, read_header=False) -> InventoryBase:
        tok_iter = _yield_schema_tokens(reader)
        # Someone else hasn't already read the header for us
        if read_header:
            schema_name, _ = next(tok_iter)
            if schema_name != cls.SCHEMA_NAME:
                raise ValueError(f"Expected schema name {schema_name!r} to be {cls.SCHEMA_NAME!r}")

        fields = cls._get_fields_dict()
        obj_dict = {}
        for key, val in tok_iter:
            if key in fields:
                field: dataclasses.Field = fields[key]
                spec = field.metadata.get("spec")
                # Not a real key, an internal var on our dataclass
                if not spec:
                    LOG.warning(f"Internal key {key!r}")
                    continue

                spec_cls = spec
                if not inspect.isclass(spec_cls):
                    spec_cls = spec_cls.__class__
                # some kind of nested structure like sale_info
                if issubclass(spec_cls, SchemaBase):
                    obj_dict[key] = spec.from_reader(reader)
                elif issubclass(spec_cls, SchemaFieldSerializer):
                    obj_dict[key] = spec.deserialize(val)
                else:
                    raise ValueError(f"Unsupported spec for {key!r}, {spec!r}")
            else:
                LOG.warning(f"Unknown key {key!r}")
        return cls._obj_from_dict(obj_dict)

    def to_writer(self, writer: StringIO):
        writer.write(f"\t{self.SCHEMA_NAME}")
        if self.SCHEMA_NAME == "permissions":
            writer.write(" 0\n")
        else:
            writer.write("\t0\n")
        writer.write("\t{\n")

        # Make sure the ID field always comes first, if there is one.
        fields_dict: Dict[str, dataclasses.Field] = {}
        if hasattr(self, "ID_ATTR"):
            fields_dict = {getattr(self, "ID_ATTR"): dataclasses.field()}
        # update()ing will put all fields that aren't yet in the dict after the ID attr.
        fields_dict.update(self._get_fields_dict())

        for field_name, field in fields_dict.items():
            spec = field.metadata.get("spec")
            # Not meant to be serialized
            if not spec:
                continue
            if field.metadata.get("llsd_only"):
                continue

            val = getattr(self, field_name)
            if val is None and not field.metadata.get("include_none"):
                continue

            spec_cls = spec
            if not inspect.isclass(spec_cls):
                spec_cls = spec_cls.__class__
            # Some kind of nested structure like sale_info
            if isinstance(val, SchemaBase):
                val.to_writer(writer)
            elif issubclass(spec_cls, SchemaFieldSerializer):
                writer.write(f"\t\t{field_name}\t{spec.serialize(val)}\n")
            else:
                raise ValueError(f"Bad inventory spec {spec!r}")
        writer.write("\t}\n")


class InventoryDifferences(NamedTuple):
    changed: List[InventoryNodeBase]
    removed: List[InventoryNodeBase]


class InventoryModel(InventoryBase):
    def __init__(self):
        self.nodes: Dict[UUID, InventoryNodeBase] = {}
        self.root: Optional[InventoryContainerBase] = None
        self.any_dirty = asyncio.Event()

    @classmethod
    def from_reader(cls, reader: StringIO, read_header=False) -> InventoryModel:
        model = cls()
        for key, value in _yield_schema_tokens(reader):
            if key == "inv_object":
                obj = InventoryObject.from_reader(reader)
                if obj is not None:
                    model.add(obj)
            elif key == "inv_category":
                cat = InventoryCategory.from_reader(reader)
                if cat is not None:
                    model.add(cat)
            elif key == "inv_item":
                item = InventoryItem.from_reader(reader)
                if item is not None:
                    model.add(item)
            else:
                LOG.warning("Unknown key {0}".format(key))
        return model

    @classmethod
    def from_llsd(cls, llsd_val: List[Dict], flavor: str = "legacy") -> Self:
        model = cls()
        for obj_dict in llsd_val:
            obj = None
            for inv_type in INVENTORY_TYPES:
                if inv_type.ID_ATTR in obj_dict:
                    if (obj := inv_type.from_llsd(obj_dict, flavor)) is not None:
                        model.add(obj)
                    break
            if obj is None:
                LOG.warning(f"Unknown object type {obj_dict!r}")
        return model

    @property
    def ordered_nodes(self) -> Iterable[InventoryNodeBase]:
        yield from self.all_containers
        yield from self.all_items

    @property
    def all_containers(self) -> Iterable[InventoryContainerBase]:
        for node in self.nodes.values():
            if isinstance(node, InventoryContainerBase):
                yield node

    @property
    def dirty_categories(self) -> Iterable[InventoryCategory]:
        for node in self.nodes:
            if isinstance(node, InventoryCategory) and node.version == InventoryCategory.VERSION_NONE:
                yield node

    @property
    def all_items(self) -> Iterable[InventoryItem]:
        for node in self.nodes.values():
            if not isinstance(node, InventoryContainerBase):
                yield node  # type: ignore

    def __eq__(self, other):
        if not isinstance(other, InventoryModel):
            return False
        return set(self.nodes.values()) == set(other.nodes.values())

    def to_writer(self, writer: StringIO):
        for node in self.ordered_nodes:
            node.to_writer(writer)

    def to_llsd(self, flavor: str = "legacy"):
        return list(node.to_llsd(flavor) for node in self.ordered_nodes)

    def add(self, node: InventoryNodeBase):
        if node.node_id in self.nodes:
            raise KeyError(f"{node.node_id} already exists in the inventory model")

        self.nodes[node.node_id] = node
        if isinstance(node, InventoryContainerBase):
            if node.parent_id == UUID.ZERO:
                self.root = node
        node.model = weakref.proxy(self)
        return node

    def update(self, node: InventoryNodeBase, update_fields: Optional[Iterable[str]] = None) -> InventoryNodeBase:
        """Update an existing node, optionally only updating specific fields"""
        if node.node_id not in self.nodes:
            raise KeyError(f"{node.node_id} not in the inventory model")

        orig_node = self.nodes[node.node_id]
        if node.__class__ != orig_node.__class__:
            raise ValueError(f"Tried to update {orig_node!r} from non-matching {node!r}")

        if not update_fields:
            # Update everything but the model parameter
            update_fields = node.get_field_names()
        for field_name in update_fields:
            setattr(orig_node, field_name, getattr(node, field_name))
        return orig_node

    def upsert(self, node: InventoryNodeBase, update_fields: Optional[Iterable[str]] = None) -> InventoryNodeBase:
        """Add or update a node"""
        if node.node_id in self.nodes:
            return self.update(node, update_fields)
        return self.add(node)

    def unlink(self, node: InventoryNodeBase, single_only: bool = False) -> Sequence[InventoryNodeBase]:
        """Unlink a node and its descendants from the tree, returning the removed nodes"""
        assert node.model == self
        if node == self.root:
            self.root = None
        unlinked = [node]
        if isinstance(node, InventoryContainerBase) and not single_only:
            for child in node.children:
                unlinked.extend(self.unlink(child))
        self.nodes.pop(node.node_id, None)
        node.model = None
        return unlinked

    def get_differences(self, other: InventoryModel) -> InventoryDifferences:
        # Includes modified things with the same ID
        changed_in_other = []
        removed_in_other = []

        other_keys = set(other.nodes.keys())
        our_keys = set(self.nodes.keys())

        # Removed
        for key in our_keys - other_keys:
            removed_in_other.append(self.nodes[key])

        # Updated
        for key in other_keys.intersection(our_keys):
            other_node = other.nodes[key]
            if other_node != self.nodes[key]:
                changed_in_other.append(other_node)

        # Added
        for key in other_keys - our_keys:
            changed_in_other.append(other.nodes[key])
        return InventoryDifferences(
            changed=changed_in_other,
            removed=removed_in_other,
        )

    def flag_if_dirty(self):
        if any(self.dirty_categories):
            self.any_dirty.set()

    def __getitem__(self, item: UUID) -> InventoryNodeBase:
        return self.nodes[item]

    def __contains__(self, item: UUID):
        return item in self.nodes

    def get(self, key: UUID) -> Optional[InventoryNodeBase]:
        return self.nodes.get(key)

    def get_category(self, key: UUID) -> InventoryCategory:
        node = self.get(key)
        if not isinstance(node, InventoryCategory):
            raise ValueError(f"{node!r} is not a category")
        return node

    def get_item(self, key: UUID) -> InventoryItem:
        node = self.get(key)
        if not isinstance(node, InventoryItem):
            raise ValueError(f"{node!r} is not an item")
        return node


@dataclasses.dataclass
class InventoryPermissions(InventoryBase):
    SCHEMA_NAME: ClassVar[str] = "permissions"

    base_mask: int = schema_field(SchemaHexInt)
    owner_mask: int = schema_field(SchemaHexInt)
    group_mask: int = schema_field(SchemaHexInt)
    everyone_mask: int = schema_field(SchemaHexInt)
    next_owner_mask: int = schema_field(SchemaHexInt)
    creator_id: UUID = schema_field(SchemaUUID)
    owner_id: UUID = schema_field(SchemaUUID)
    last_owner_id: UUID = schema_field(SchemaUUID)
    group_id: UUID = schema_field(SchemaUUID)
    # Nothing actually cares about this, but it could be there.
    # It's kind of redundant since it just means owner_id == NULL_KEY && group_id != NULL_KEY.
    is_owner_group: Optional[int] = schema_field(SchemaInt, default=None, llsd_only=True)

    @classmethod
    def make_default(cls) -> Self:
        return cls(
            base_mask=0xFFffFFff,
            owner_mask=0xFFffFFff,
            group_mask=0,
            everyone_mask=0,
            next_owner_mask=0x82000,
            creator_id=UUID.ZERO,
            owner_id=UUID.ZERO,
            last_owner_id=UUID.ZERO,
            group_id=UUID.ZERO,
            is_owner_group=None
        )


@dataclasses.dataclass
class InventorySaleInfo(InventoryBase):
    SCHEMA_NAME: ClassVar[str] = "sale_info"

    sale_type: SaleType = schema_field(SchemaEnumField(SaleType))
    sale_price: int = schema_field(SchemaInt)

    @classmethod
    def make_default(cls) -> Self:
        return cls(sale_type=SaleType.NOT, sale_price=10)


class _HasBaseNodeAttrs(abc.ABC):
    """
    Only exists so that we can assert that all subclasses should have this without forcing
    a particular serialization order, as would happen if this was present on InventoryNodeBase.
    """
    name: str
    type: AssetType


@dataclasses.dataclass
class InventoryNodeBase(InventoryBase, _HasBaseNodeAttrs):
    ID_ATTR: ClassVar[str]

    parent_id: Optional[UUID] = schema_field(SchemaUUID)

    model: Optional[InventoryModel] = dataclasses.field(
        default=None, init=False, hash=False, compare=False, repr=False
    )

    @classmethod
    def get_field_names(cls) -> Set[str]:
        return set(cls._get_fields_dict().keys()) - {"model"}

    @property
    def node_id(self) -> UUID:
        return getattr(self, self.ID_ATTR)

    @node_id.setter
    def node_id(self, val: UUID):
        setattr(self, self.ID_ATTR, val)

    @property
    def parent(self) -> Optional[InventoryContainerBase]:
        return self.model.nodes.get(self.parent_id)

    def unlink(self) -> Sequence[InventoryNodeBase]:
        return self.model.unlink(self)

    @classmethod
    def _obj_from_dict(cls, obj_dict):
        # Bad entry, ignore
        # TODO: Check on these. might be symlinks or something.
        if obj_dict.get("type") == "-1":
            LOG.warning(f"Skipping bad object with type == -1: {obj_dict!r}")
            return None
        return super()._obj_from_dict(obj_dict)

    def __hash__(self):
        return hash(self.node_id)

    def __iter__(self) -> Iterator[InventoryNodeBase]:
        return iter(())

    def __contains__(self, item) -> bool:
        return item in tuple(self)


@dataclasses.dataclass
class InventoryContainerBase(InventoryNodeBase):
    type: AssetType = schema_field(SchemaEnumField(AssetType))

    @property
    def children(self) -> Sequence[InventoryNodeBase]:
        return tuple(
            x for x in self.model.nodes.values()
            if x.parent_id == self.node_id
        )

    @property
    def descendents(self) -> List[InventoryNodeBase]:
        new_children: List[InventoryNodeBase] = [self]
        descendents = []
        while new_children:
            to_check = new_children[:]
            new_children.clear()
            for obj in to_check:
                if isinstance(obj, InventoryContainerBase):
                    for child in obj.children:
                        if child in descendents:
                            continue
                        new_children.append(child)
                        descendents.append(child)
                else:
                    if obj not in descendents:
                        descendents.append(obj)
        return descendents

    def __getitem__(self, item: Union[int, str]) -> InventoryNodeBase:
        if isinstance(item, int):
            return self.children[item]

        for child in self.children:
            if child.name == item:
                return child
        raise KeyError(f"{item!r} not found in children")

    def __iter__(self) -> Iterator[InventoryNodeBase]:
        return iter(self.children)

    def get_or_create_subcategory(self, name: str) -> InventoryCategory:
        for child in self:
            if child.name == name and isinstance(child, InventoryCategory):
                return child
        child = InventoryCategory(
            name=name,
            cat_id=UUID.random(),
            parent_id=self.node_id,
            type=AssetType.CATEGORY,
            pref_type=FolderType.NONE,
            owner_id=getattr(self, 'owner_id', UUID.ZERO),
            version=1,
        )
        self.model.add(child)
        return child

    # So autogenerated __hash__ doesn't kill our inherited one
    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryObject(InventoryContainerBase):
    SCHEMA_NAME: ClassVar[str] = "inv_object"
    ID_ATTR: ClassVar[str] = "obj_id"

    obj_id: UUID = schema_field(SchemaUUID)
    name: str = schema_field(SchemaMultilineStr)
    metadata: Optional[Dict[str, Any]] = schema_field(SchemaLLSD, default=None, include_none=True)

    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryCategory(InventoryContainerBase):
    ID_ATTR: ClassVar[str] = "cat_id"
    # AIS calls this something else...
    ID_ATTR_AIS: ClassVar[str] = "category_id"
    SCHEMA_NAME: ClassVar[str] = "inv_category"
    VERSION_NONE: ClassVar[int] = -1

    cat_id: UUID = schema_field(SchemaUUID)
    pref_type: FolderType = schema_field(SchemaEnumField(FolderType), llsd_name="preferred_type")
    name: str = schema_field(SchemaMultilineStr)
    owner_id: Optional[UUID] = schema_field(SchemaUUID, default=None)
    version: int = schema_field(SchemaInt, default=VERSION_NONE, llsd_only=True)
    metadata: Optional[Dict[str, Any]] = schema_field(SchemaLLSD, default=None, include_none=False)

    def to_folder_data(self) -> Block:
        return Block(
            "FolderData",
            FolderID=self.cat_id,
            ParentID=self.parent_id,
            CallbackID=0,
            Type=self.pref_type,
            Name=self.name,
        )

    @classmethod
    def from_folder_data(cls, block: Block):
        return cls(
            cat_id=block["FolderID"],
            parent_id=block["ParentID"],
            pref_type=block["Type"],
            name=block["Name"],
            type=AssetType.CATEGORY,
        )

    @classmethod
    def from_llsd(cls, inv_dict: Dict, flavor: str = "legacy") -> Self:
        if flavor == "ais" and "type" not in inv_dict:
            inv_dict = inv_dict.copy()
            inv_dict["type"] = AssetType.CATEGORY
        return super().from_llsd(inv_dict, flavor)

    def to_llsd(self, flavor: str = "legacy"):
        payload = super().to_llsd(flavor)
        if flavor == "ais":
            # AIS already knows the inventory type is category
            payload.pop("type", None)
        return payload

    @classmethod
    def _get_fields_dict(cls, llsd_flavor: Optional[str] = None):
        fields = super()._get_fields_dict(llsd_flavor)
        if llsd_flavor == "ais":
            # These have different names though
            fields["type_default"] = fields.pop("preferred_type")
            fields["agent_id"] = fields.pop("owner_id")
            fields["category_id"] = fields.pop("cat_id")
        return fields

    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryItem(InventoryNodeBase):
    SCHEMA_NAME: ClassVar[str] = "inv_item"
    ID_ATTR: ClassVar[str] = "item_id"

    item_id: UUID = schema_field(SchemaUUID)
    permissions: InventoryPermissions = schema_field(InventoryPermissions)
    asset_id: Optional[UUID] = schema_field(SchemaUUID, default=None)
    shadow_id: Optional[UUID] = schema_field(SchemaUUID, default=None)
    type: Optional[AssetType] = schema_field(SchemaEnumField(AssetType), default=None)
    inv_type: Optional[InventoryType] = schema_field(SchemaEnumField(InventoryType), default=None)
    flags: Optional[int] = schema_field(SchemaFlagField, default=None)
    sale_info: Optional[InventorySaleInfo] = schema_field(InventorySaleInfo, default=None)
    name: Optional[str] = schema_field(SchemaMultilineStr, default=None)
    desc: Optional[str] = schema_field(SchemaMultilineStr, default=None)
    metadata: Optional[Dict[str, Any]] = schema_field(SchemaLLSD, default=None, include_none=True)
    """Specifically for script metadata, generally just experience info"""
    thumbnail: Optional[Dict[str, Any]] = schema_field(SchemaLLSD, default=None, include_none=False)
    """Generally just a dict with the thumbnail UUID in it"""
    creation_date: Optional[dt.datetime] = schema_field(SchemaDate, llsd_name="created_at", default=None)

    __hash__ = InventoryNodeBase.__hash__

    @property
    def true_asset_id(self) -> UUID:
        if self.asset_id is not None:
            return self.asset_id
        return self.shadow_id ^ MAGIC_ID

    def to_inventory_data(self, block_name: str = "InventoryData") -> Block:
        return Block(
            block_name,
            ItemID=self.item_id,
            FolderID=self.parent_id,
            CallbackID=0,
            CreatorID=self.permissions.creator_id,
            OwnerID=self.permissions.owner_id,
            GroupID=self.permissions.group_id,
            BaseMask=self.permissions.base_mask,
            OwnerMask=self.permissions.owner_mask,
            GroupMask=self.permissions.group_mask,
            EveryoneMask=self.permissions.everyone_mask,
            NextOwnerMask=self.permissions.next_owner_mask,
            GroupOwned=self.permissions.owner_id == UUID.ZERO and self.permissions.group_id != UUID.ZERO,
            AssetID=self.true_asset_id,
            Type=self.type,
            InvType=self.inv_type,
            Flags=self.flags,
            SaleType=self.sale_info.sale_type,
            SalePrice=self.sale_info.sale_price,
            Name=self.name,
            Description=self.desc,
            CreationDate=SchemaDate.to_llsd(self.creation_date, "legacy"),
            # Meaningless here
            CRC=secrets.randbits(32),
        )

    @classmethod
    def from_inventory_data(cls, block: Block):
        return cls(
            item_id=block["ItemID"],
            # Might be under one of two names
            parent_id=block.get("ParentID", block["FolderID"]),
            permissions=InventoryPermissions(
                creator_id=block["CreatorID"],
                owner_id=block["OwnerID"],
                # Unknown, not sent in this schema
                last_owner_id=block.get("LastOwnerID", UUID.ZERO),
                group_id=block["GroupID"],
                base_mask=block["BaseMask"],
                owner_mask=block["OwnerMask"],
                group_mask=block["GroupMask"],
                everyone_mask=block["EveryoneMask"],
                next_owner_mask=block["NextOwnerMask"],
            ),
            # May be missing in UpdateInventoryItem
            asset_id=block.get("AssetID"),
            type=AssetType(block["Type"]),
            inv_type=InventoryType(block["InvType"]),
            flags=block["Flags"],
            sale_info=InventorySaleInfo(
                sale_type=SaleType(block["SaleType"]),
                sale_price=block["SalePrice"],
            ),
            name=block["Name"],
            desc=block["Description"],
            creation_date=SchemaDate.from_llsd(block["CreationDate"], "legacy"),
        )

    def to_llsd(self, flavor: str = "legacy"):
        val = super().to_llsd(flavor=flavor)
        if flavor == "ais":
            # There's little chance this differs from owner ID, just place it.
            val["agent_id"] = val["permissions"]["owner_id"]
            if val["type"] == AssetType.LINK:
                # For link items, there is no asset, only a linked ID.
                val["linked_id"] = val.pop("asset_id")
                # These don't exist either
                val.pop("permissions", None)
                val.pop("sale_info", None)
        return val

    @classmethod
    def from_llsd(cls, inv_dict: Dict, flavor: str = "legacy") -> Self:
        if flavor == "ais" and "linked_id" in inv_dict:
            # Links get represented differently than other items for whatever reason.
            # This is incredibly annoying, under *NIX there's nothing really special about symlinks.
            inv_dict = inv_dict.copy()
            # Fill this in since it needs to be there
            if "permissions" not in inv_dict:
                inv_dict["permissions"] = InventoryPermissions(
                    base_mask=0xFFffFFff,
                    owner_mask=0xFFffFFff,
                    group_mask=0xFFffFFff,
                    everyone_mask=0,
                    next_owner_mask=0xFFffFFff,
                    creator_id=UUID.ZERO,
                    owner_id=UUID.ZERO,
                    last_owner_id=UUID.ZERO,
                    group_id=UUID.ZERO,
                ).to_llsd("ais")
            if "sale_info" not in inv_dict:
                inv_dict["sale_info"] = InventorySaleInfo(
                    sale_type=SaleType.NOT,
                    sale_price=0,
                ).to_llsd("ais")
            if "type" not in inv_dict:
                inv_dict["type"] = AssetType.LINK

            # In the context of symlinks, asset id means linked item ID.
            # This is also how indra stores symlinks. Why the asymmetry in AIS if none of the
            # consumers actually want it? Who knows.
            inv_dict["asset_id"] = inv_dict.pop("linked_id")
        return super().from_llsd(inv_dict, flavor)


INVENTORY_TYPES: Tuple[Type[InventoryNodeBase], ...] = (InventoryCategory, InventoryObject, InventoryItem)



"""
Validates that serialize(deserialize(packet)) == packet for any packet
that passes through the proxy. Useful for ensuring that serializers don't
change the meaning of a message, and that all of the viewer's quirks are
faithfully reproduced.
"""

import copy
import itertools
import logging
from typing import *

from hippolyzer.lib.base.message.msgtypes import PacketLayout
from hippolyzer.lib.base import serialization as se
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import UDPPacket
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import SessionManager, Session

LOG = logging.getLogger(__name__)


class SerializationSanityChecker(BaseAddon):
    def __init__(self):
        self.serializer = UDPMessageSerializer()
        self.deserializer = UDPMessageDeserializer()

    def handle_proxied_packet(self, session_manager: SessionManager, packet: UDPPacket,
                              session: Optional[Session], region: Optional[ProxiedRegion]):
        # Well this doesn't even parse as a message, can't do anything about it.
        try:
            message = self.deserializer.deserialize(packet.data)
        except:
            LOG.error(f"Received unparseable message from {packet.src_addr!r}: {packet.data!r}")
            return
        try:
            message.ensure_parsed()
        except:
            LOG.exception(f"Exception during {message.name} message validation pre-parsing")
            return

        try:
            # We already know the message won't match if the serializers don't roundtrip.
            if message and self._roundtrip_var_serializers(message):
                ser = self.serializer.serialize(message)
                # LL's ObjectUpdate specifically randomly uses inefficient zero-coding
                # which is hard to reproduce. It means the same thing when decompressed,
                # so just expand both and compare. Technically this incorrectly expands the
                # acks too, but shouldn't matter because they should be the same in both.
                if message.name == "ObjectUpdate" and message.zerocoded:
                    orig_body = self.deserializer.zero_code_expand(packet.data[PacketLayout.PHL_NAME:])
                    ser_body = self.deserializer.zero_code_expand(ser[PacketLayout.PHL_NAME:])
                    matches = orig_body == ser_body
                else:
                    matches = packet.data == ser

                if not matches:
                    direction = "Out" if packet.outgoing else "In"
                    LOG.error("%s: %d %s\n%r != %r" %
                              (direction, message.packet_id, message.name, packet.data, ser))
        except:
            LOG.exception(f"Exception during message validation:\n{message!r}")

    def _roundtrip_var_serializers(self, message: Message):
        for block in itertools.chain(*message.blocks.values()):
            for var_name in block.vars.keys():
                orig_val = block[var_name]
                try:
                    orig_serializer = block.get_serializer(var_name)
                except KeyError:
                    # Don't have a serializer, onto the next field
                    continue
                # need to copy the serializer since we're going to replace a member function
                serializer: se.BaseSubfieldSerializer = copy.copy(orig_serializer)

                # Keep track of what got serialized at what position
                member_positions = []

                def _serialize_template(template, val):
                    writer = se.MemberTrackingBufferWriter(serializer.ENDIANNESS)
                    writer.write(template, val)
                    member_positions.clear()
                    member_positions.extend(writer.member_positions)
                    return writer.copy_buffer()

                serializer._serialize_template = _serialize_template
                try:
                    deser = serializer.deserialize(block, orig_val)
                except:
                    LOG.error(f"Exploded in deserializer for {message.name}.{block.name}.{var_name}")
                    raise

                # For now we consider returning UNSERIALIZABLE to be acceptable.
                # We should probably consider raising instead of returning that.
                if deser is se.UNSERIALIZABLE:
                    continue

                try:
                    new_val = serializer.serialize(block, deser)
                except:
                    LOG.error(f"Exploded in serializer for {message.name}.{block.name}.{var_name}")
                    raise

                if orig_val != new_val:
                    # OpenSim will put an extra NUL at the end of TEs with material fields
                    # whereas the viewer and SL just use EOF rather than explicit NUL to signal
                    # the end of the exception cases for the last field in a TE.
                    # OpenSim's behaviour isn't incorrect, but we're not going to reproduce it.
                    if var_name == "TextureEntry" and orig_val[:-1] == new_val and orig_val[-1] == 0:
                        continue
                    LOG.error("%d %s.%s.%s\n%r != %r" %
                              (message.packet_id, message.name, block.name, var_name, orig_val, new_val))
                    # This was templated, we can dig into which member mismatched
                    if member_positions:
                        # find the mismatch index
                        i = 0
                        bytes_zipped = itertools.zip_longest(orig_val, new_val, fillvalue=object())
                        for i, (old_byte, new_byte) in enumerate(bytes_zipped):
                            if old_byte != new_byte:
                                break
                        LOG.error(f"Mismatch at {i}, {member_positions!r}")
                    return False
        return True


addons = [SerializationSanityChecker()]



from typing import *

import asyncio
import unittest
from unittest import mock

from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.voice.client import VoiceClient
from hippolyzer.lib.voice.connection import VivoxConnection


def _make_transport(write_func):
    transport = mock.Mock()
    transport.write.side_effect = write_func
    transport.is_closing.return_value = False
    return transport


def _make_protocol(transport: Any):
    protocol = mock.Mock(transport=transport)
    protocol._drain_helper = mock.AsyncMock()
    return protocol


class TestVivoxConnection(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self._writer_buf = bytearray()
        self._transport = _make_transport(self._writer_buf.extend)
        self._protocol = _make_protocol(self._transport)
        self.reader = asyncio.StreamReader()
        self.writer = asyncio.StreamWriter(self._transport, self._protocol, self.reader, asyncio.get_event_loop())
        self.vivox_connection = VivoxConnection(self.reader, self.writer)

    async def test_read_request(self):
        self.reader.feed_data(
            b'<Request requestId="foobar" action="Aux.GetRenderDevices.1"><Foo>1</Foo></Request>\n\n\n'
        )
        self.reader.feed_eof()
        msg_type, msg_action, request_id, body = await self.vivox_connection.read_message()
        self.assertEqual("Request", msg_type)
        self.assertEqual("Aux.GetRenderDevices.1", msg_action)
        self.assertEqual("foobar", request_id)
        self.assertDictEqual({"Foo": "1"}, body)

    async def test_read_response(self):
        self.reader.feed_data(
            b'<Response requestId="foobar" action="Connector.SetLocalMicVolume.1"><ReturnCode>0</ReturnCode>'
            b'<Results><StatusCode>0</StatusCode><StatusString /></Results>'
            b'<InputXml><Request/></InputXml></Response>\n\n\n'
        )
        self.reader.feed_eof()
        msg_type, msg_action, request_id, body = await self.vivox_connection.read_message()
        self.assertEqual("Response", msg_type)
        self.assertEqual("Connector.SetLocalMicVolume.1", msg_action)
        self.assertEqual("foobar", request_id)
        self.assertDictEqual(
            {'ReturnCode': 0, 'Results': {'StatusCode': '0', 'StatusString': None}},
            body,
        )

    async def test_read_event(self):
        self.reader.feed_data(
            b'<Event type="MediaStreamUpdatedEvent"><SessionGroupHandle>4</SessionGroupHandle><SessionHandle>7'
            b'</SessionHandle><StatusCode>0</StatusCode><StatusString/>'
            b'<State>6</State><StateDescription>Connecting</StateDescription><Incoming>false</Incoming>'
            b'<DurableMediaId/></Event>\n\n\n'
        )
        self.reader.feed_eof()
        msg_type, msg_action, request_id, body = await self.vivox_connection.read_message()
        self.assertEqual("Event", msg_type)
        self.assertEqual("MediaStreamUpdatedEvent", msg_action)
        self.assertEqual(None, request_id)
        self.assertDictEqual(
            {
                'DurableMediaId': None,
                'Incoming': 'false',
                'SessionGroupHandle': '4',
                'SessionHandle': '7',
                'State': '6',
                'StateDescription': 'Connecting',
                'StatusCode': '0',
                'StatusString': None,
            },
            body,
        )

    async def test_read_messages(self):
        self.reader.feed_data(
            b'<Request requestId="foobar" action="Aux.GetRenderDevices.1"><Foo>1</Foo></Request>\n\n\n'
            b'<Request requestId="quux" action="Aux.GetRenderDevices.1"><Foo>1</Foo></Request>\n\n\n'
        )
        self.reader.feed_eof()

        i = 0
        async for msg in self.vivox_connection.read_messages():
            if i == 0:
                self.assertEqual("foobar", msg.request_id)
            else:
                self.assertEqual("quux", msg.request_id)

            self.assertEqual("Request", msg.type)
            self.assertEqual("Aux.GetRenderDevices.1", msg.name)
            self.assertDictEqual({"Foo": "1"}, msg.data)
            i += 1

    async def test_send_message(self):
        await self.vivox_connection.send_request("foo", "bar", {"baz": 1})
        self.assertEqual(
            b'<Request requestId="foo" action="bar"><baz>1</baz></Request>\n\n\n',
            self._writer_buf
        )


class TestVoiceClient(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self._client_transport = _make_transport(
            lambda *args: asyncio.get_event_loop().call_soon(self.server_reader.feed_data, *args)
        )
        self._client_protocol = _make_protocol(self._client_transport)
        self.client_reader = asyncio.StreamReader()
        self.client_writer = asyncio.StreamWriter(
            self._client_transport,
            self._client_protocol,
            self.client_reader,
            asyncio.get_event_loop()
        )

        self._server_transport = _make_transport(
            lambda *args: asyncio.get_event_loop().call_soon(self.client_reader.feed_data, *args)
        )
        self._server_protocol = _make_protocol(self._server_transport)
        self.server_reader = asyncio.StreamReader()
        self.server_writer = asyncio.StreamWriter(
            self._server_transport,
            self._server_protocol,
            self.server_reader,
            asyncio.get_event_loop()
        )

        self.client_connection = VivoxConnection(self.client_reader, self.client_writer)
        self.server_connection = VivoxConnection(self.server_reader, self.server_writer)
        self.client = VoiceClient("127.0.0.1", 0)
        self.client.vivox_conn = self.client_connection

        def _make_request_id():
            _make_request_id.i += 1
            return str(_make_request_id.i)

        _make_request_id.i = 0

        self.client._make_request_id = _make_request_id

    async def _expect_message(self, name: str):
        msg = await self.server_connection.read_message()
        self.assertEqual(name, msg.name)
        return msg

    async def _handle_message(self, name: str):
        msg = await self._expect_message(name)
        await self.server_connection.send_response(msg.request_id, msg.name, {
            "ReturnCode": 0,
            "Results": {}
        })
        return msg

    async def _do_connector_setup(self):
        async def _serve_connector_setup():
            await self.server_connection.send_event(
                "VoiceServiceConnectionStateChangedEvent",
                {
                    "Connected": 1,
                    "Platform": "Linux",
                    "Version": 1,
                    "DataDirectory": "/tmp/whatever",
                }
            )
            msg = await self.server_connection.read_message()
            self.assertEqual('Aux.GetCaptureDevices.1', msg.name)
            await self.server_connection.send_response(msg.request_id, msg.name, {
                "ReturnCode": 0,
                "Results": {
                    "StatusCode": 0,
                    "StatusString": None,
                    "CaptureDevices": []
                }
            })

            msg = await self.server_connection.read_message()
            self.assertEqual('Aux.GetRenderDevices.1', msg.name)
            await self.server_connection.send_response(msg.request_id, msg.name, {
                "ReturnCode": 0,
                "Results": {
                    "StatusCode": 0,
                    "StatusString": None,
                    "RenderDevices": []
                }
            })
            await self._handle_message("Connector.MuteLocalSpeaker.1")
            await self._handle_message("Connector.SetLocalSpeakerVolume.1")
            await self._handle_message("Connector.MuteLocalMic.1")
            await self._handle_message("Connector.SetLocalMicVolume.1")

            msg = await self.server_connection.read_message()
            self.assertEqual('Connector.Create.1', msg.name)
            await self.server_connection.send_response(msg.request_id, msg.name, {
                "ReturnCode": 0,
                "Results": {
                    "ConnectorHandle": 2,
                }
            })

        serve_connector_task = asyncio.create_task(_serve_connector_setup())
        await asyncio.wait_for(serve_connector_task, 0.5)
        await asyncio.wait_for(self.client.ready.wait(), 0.5)

    async def _do_login(self):
        async def _serve_login():
            msg = await self._expect_message("Account.Login.1")
            self.assertEqual("foo", msg.data["AccountName"])
            await self.server_connection.send_event("AccountLoginStateChangeEvent", {
                "AccountHandle": 2,
                "StatusCode": 200,
                "StatusString": "OK",
                "State": 1,
            })
            await self.server_connection.send_response(msg.request_id, msg.name, {
                "ReturnCode": 0,
                "Results": {
                    "StatusCode": 0,
                    "StatusString": None,
                    "AccountHandle": 2,
                    "DisplayName": "foo",
                    "Uri": "uri:baz@foo",
                }
            })

        login_task = asyncio.create_task(_serve_login())

        await asyncio.wait_for(self.client.login("foo", "bar"), 0.5)
        await asyncio.wait_for(login_task, 0.5)

    async def _join_session(self):
        async def _serve_session():
            await self._handle_message("Session.Create.1")
            await self.server_connection.send_event("SessionAddedEvent", {
                "SessionHandle": 4,
                "SessionGroupHandle": 5,
            })
            await self.server_connection.send_event("ParticipantAddedEvent", {
                "ParticipantUri": "uri:baz@foo",
            })

        serve_session_task = asyncio.create_task(_serve_session())

        await asyncio.wait_for(self.client.join_session("uri:foo@bar", region_handle=256), 0.5)
        self.assertIn("uri:baz@foo", self.client.participants)

        await asyncio.wait_for(serve_session_task, 0.5)

    async def test_create_connector(self):
        await self._do_connector_setup()

    async def test_login(self):
        await self._do_connector_setup()
        await self._do_login()

    async def test_create_session(self):
        await self._do_connector_setup()
        await self._do_login()
        await self._join_session()

    async def test_set_position(self):
        await self._do_connector_setup()
        await self._do_login()
        await self._join_session()
        handle_3d_pos_task = asyncio.create_task(self._handle_message("Session.Set3DPosition.1"))
        await self.client.set_region_3d_pos(Vector3(1, 2, 3))
        msg = await handle_3d_pos_task
        self.assertDictEqual(
            {'X': '1.0', 'Y': '3.0', 'Z': '-258.0'},
            msg.data["SpeakerPosition"]["Position"],
        )
        self.assertAlmostEqual(self.client.region_pos.X, 1.0)
        self.assertAlmostEqual(self.client.region_pos.Y, 2.0)
        self.assertAlmostEqual(self.client.region_pos.Z, 3.0)

        self.assertAlmostEqual(self.client.global_pos.X, 1.0)
        self.assertAlmostEqual(self.client.global_pos.Y, 3.0)
        self.assertAlmostEqual(self.client.global_pos.Z, -258.0)



import glob

import setuptools  # noqa

import os
import shutil
from distutils.core import Command
from importlib.metadata import version
from pathlib import Path

from cx_Freeze import setup, Executable

# We don't need any of these and they make the archive huge.
TO_DELETE = [
    "lib/PySide6/Qt6DRender.pyd",
    "lib/PySide6/Qt63DRender.dll",
    "lib/PySide6/Qt6Charts.dll",
    "lib/PySide6/Qt6Location.dll",
    "lib/PySide6/Qt6Pdf.dll",
    "lib/PySide6/Qt6Quick.dll",
    "lib/PySide6/Qt6WebEngineCore.dll",
    "lib/PySide6/QtCharts.pyd",
    "lib/PySide6/QtMultimedia.pyd",
    "lib/PySide6/QtOpenGLFunctions.pyd",
    "lib/PySide6/QtOpenGLFunctions.pyi",
    "lib/PySide6/d3dcompiler_47.dll",
    "lib/PySide6/opengl32sw.dll",
    "lib/PySide6/lupdate.exe",
    "lib/PySide6/translations",
    "lib/aiohttp/_find_header.c",
    "lib/aiohttp/_frozenlist.c",
    "lib/aiohttp/_helpers.c",
    "lib/aiohttp/_http_parser.c",
    "lib/aiohttp/_http_writer.c",
    "lib/aiohttp/_websocket.c",
    # Improve this to work with different versions.
    "lib/aiohttp/python3*.dll",
    "lib/lazy_object_proxy/python3*.dll",
    "lib/lxml/python3*.dll",
    "lib/markupsafe/python3*.dll",
    "lib/multidict/python3*.dll",
    "lib/numpy/core/python3*.dll",
    "lib/numpy/fft/python3*.dll",
    "lib/numpy/linalg/python3*.dll",
    "lib/numpy/random/python3*.dll",
    "lib/python3*.dll",
    "lib/recordclass/python3*.dll",
    "lib/regex/python3*.dll",
    "lib/test",
    "lib/yarl/python3*.dll",
]

COPY_TO_ZIP = [
    "LICENSE.txt",
    "README.md",
    "NOTICE.md",
    # Must have been generated with pip-licenses before. Many dependencies
    # require their license to be distributed with their binaries.
    "lib_licenses.txt",
]


BASE_DIR = Path(__file__).parent.absolute()


class FinalizeCXFreezeCommand(Command):
    description = "Prepare cx_Freeze build dirs and create a zip"
    user_options = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self):
        (BASE_DIR / "dist").mkdir(exist_ok=True)
        for path in (BASE_DIR / "build").iterdir():
            if path.name.startswith("exe.") and path.is_dir():
                for cleanse_suffix in TO_DELETE:
                    cleanse_path = path / cleanse_suffix
                    for globbed in glob.glob(str(cleanse_path)):
                        shutil.rmtree(globbed, ignore_errors=True)
                        try:
                            os.unlink(globbed)
                        except:
                            pass
                for to_copy in COPY_TO_ZIP:
                    shutil.copy(BASE_DIR / to_copy, path / to_copy)
                shutil.copytree(BASE_DIR / "addon_examples", path / "addon_examples")
                zip_path = BASE_DIR / "dist" / path.name
                shutil.make_archive(zip_path, "zip", path)


options = {
    "build_exe": {
        "packages": [
            "passlib",
            "_cffi_backend",
            "hippolyzer",
            "mitmproxy_windows",
        ],
        # exclude packages that are not really needed
        "excludes": [
            "tkinter",
        ],
        "include_msvcr": True,
    }
}

executables = [
    Executable(
        "hippolyzer/apps/proxy_gui.py",
        base=None,
        target_name="hippolyzer_gui"
    ),
]

setup(
    name="hippolyzer_gui",
    version=version("hippolyzer"),
    description="Hippolyzer GUI",
    options=options,
    executables=executables,
    cmdclass={
        "finalize_cxfreeze": FinalizeCXFreezeCommand,
    }
)



import random

from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class SimulatePacketLossAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        # Messing with these may kill your circuit
        if message.name in {"PacketAck", "StartPingCheck", "CompletePingCheck", "UseCircuitCode",
                            "CompleteAgentMovement", "AgentMovementComplete"}:
            return
        # Simulate 30% packet loss
        if random.random() > 0.7:
            # Do nothing, drop this packet on the floor
            return True
        return


addons = [SimulatePacketLossAddon()]



"""
Helper for making deformer anims. This could have a GUI I guess.
"""
import dataclasses
from typing import *

import numpy as np
import transformations

from hippolyzer.lib.base.datatypes import Vector3, Quaternion, UUID
from hippolyzer.lib.base.llanim import Joint, Animation, PosKeyframe, RotKeyframe
from hippolyzer.lib.base.mesh import MeshAsset, SegmentHeaderDict, SkinSegmentDict, LLMeshSerializer
from hippolyzer.lib.base.serialization import BufferWriter
from hippolyzer.lib.proxy.addon_utils import show_message, BaseAddon, SessionProperty
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.commands import handle_command, Parameter
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session

import local_anim
# We require any addons from local_anim to be loaded, and we want
# our addon to be reloaded whenever local_anim changes.
AddonManager.hot_reload(local_anim, require_addons_loaded=True)


@dataclasses.dataclass
class DeformerJoint:
    pos: Optional[Vector3] = None
    rot: Optional[Quaternion] = None


def build_deformer(joints: Dict[str, DeformerJoint]) -> bytes:
    anim = Animation(
        major_version=1,
        minor_version=0,
        base_priority=5,
        duration=1.0,
        loop_out_point=1.0,
        loop=True,
    )

    for joint_name, joint in joints.items():
        if not any((joint.pos, joint.rot)):
            continue
        anim.joints[joint_name] = Joint(
            priority=5,
            rot_keyframes=[RotKeyframe(time=0.0, rot=joint.rot)] if joint.rot else [],
            pos_keyframes=[PosKeyframe(time=0.0, pos=joint.pos)] if joint.pos else [],
        )
    return anim.to_bytes()


def build_mesh_deformer(joints: Dict[str, DeformerJoint]) -> bytes:
    skin_seg = SkinSegmentDict(
        joint_names=[],
        bind_shape_matrix=identity_mat4(),
        inverse_bind_matrix=[],
        alt_inverse_bind_matrix=[],
        pelvis_offset=0.0,
        lock_scale_if_joint_position=False
    )
    for joint_name, joint in joints.items():
        # We can only represent joint translations, ignore this joint if it doesn't have any.
        if not joint.pos:
            continue
        skin_seg['joint_names'].append(joint_name)
        # Inverse bind matrix isn't actually used, so we can just give it a placeholder value of the
        # identity mat4. This might break things in weird ways because the matrix isn't actually sensible.
        skin_seg['inverse_bind_matrix'].append(identity_mat4())
        # Create a flattened mat4 that only has a translation component of our joint pos
        # The viewer ignores any other component of these matrices so no point putting shear
        # or perspective or whatever :)
        joint_mat4 = pos_to_mat4(joint.pos)
        # Ask the viewer to override this joint's usual parent-relative position with our matrix
        skin_seg['alt_inverse_bind_matrix'].append(joint_mat4)

    # Make a dummy mesh and shove our skin segment onto it. None of the tris are rigged, so the
    # viewer will freak out and refuse to display the tri, only the joint translations will be used.
    # Supposedly a mesh with a `skin` segment but no weights on the material should just result in an
    # effectively unrigged material, but that's not the case. Oh well.
    mesh = MeshAsset.make_triangle()
    mesh.header['skin'] = SegmentHeaderDict(offset=0, size=0)
    mesh.segments['skin'] = skin_seg

    writer = BufferWriter("!")
    writer.write(LLMeshSerializer(), mesh)
    return writer.copy_buffer()


def identity_mat4() -> List[float]:
    """
    Return an "Identity" mat4

    Effectively represents a transform of no rot, no translation, no shear, no perspective
    and scaling by 1.0 on every axis.
    """
    return list(np.identity(4).flatten('F'))


def pos_to_mat4(pos: Vector3) -> List[float]:
    """Convert a position Vector3 to a Translation Mat4"""
    return list(transformations.compose_matrix(translate=tuple(pos)).flatten('F'))


class DeformerAddon(BaseAddon):
    deform_joints: Dict[str, DeformerJoint] = SessionProperty(dict)

    @handle_command()
    async def save_deformer(self, _session: Session, _region: ProxiedRegion):
        filename = await AddonManager.UI.save_file(filter_str="SL Anim (*.anim)")
        if not filename:
            return
        with open(filename, "wb") as f:
            f.write(build_deformer(self.deform_joints))

    # `sep=None` makes `coord` greedy, taking the rest of the message
    @handle_command(
        joint_name=str,
        coord_type=str,
        coord=Parameter(Vector3.parse, sep=None),
    )
    async def set_deformer_joint(self, session: Session, region: ProxiedRegion,
                                 joint_name: str, coord_type: str, coord: Vector3):
        """
        Set a coordinate for a joint in the deformer

        Example:
            set_deformer_joint mNeck pos <0, 0, 0.5>
            set_deformer_joint mNeck rot <0, 180, 0>
        """
        joint_data = self.deform_joints.setdefault(joint_name, DeformerJoint())

        if coord_type == "pos":
            joint_data.pos = coord
        elif coord_type == "rot":
            joint_data.rot = Quaternion.from_euler(*coord, degrees=True)
        else:
            show_message(f"Unknown deformer component {coord_type}")
            return
        self._reapply_deformer(session, region)

    @handle_command()
    async def stop_deforming(self, session: Session, region: ProxiedRegion):
        """Disable any active deformer, may have to reset skeleton manually"""
        self.deform_joints.clear()
        self._reapply_deformer(session, region)

    def _reapply_deformer(self, session: Session, region: ProxiedRegion):
        anim_data = None
        if self.deform_joints:
            anim_data = build_deformer(self.deform_joints)
        local_anim.LocalAnimAddon.apply_local_anim(session, region, "deformer_addon", anim_data)

    def handle_rlv_command(self, session: Session, region: ProxiedRegion, source: UUID,
                           behaviour: str, options: List[str], param: str):
        # An object in-world can also tell the client how to deform itself via
        # RLV-style commands.

        # We only handle commands
        if param != "force":
            return

        if behaviour == "stop_deforming":
            self.deform_joints.clear()
        elif behaviour == "deform_joints":
            self.deform_joints.clear()
            for joint_data in options:
                joint_split = joint_data.split("|")
                pos = Vector3(*joint_split[1].split("/")) if joint_split[1] else None
                rot = Quaternion(*joint_split[2].split("/")) if joint_split[2] else None
                self.deform_joints[joint_split[0]] = DeformerJoint(pos=pos, rot=rot)
        else:
            return

        self._reapply_deformer(session, region)
        return True

    @handle_command()
    async def save_deformer_as_mesh(self, _session: Session, _region: ProxiedRegion):
        """
        Export the deformer as a crafted rigged mesh rather than an animation

        Mesh deformers have the advantage that they don't cause your joints to "stick"
        like animations do when using animations with pos keyframes.
        """
        filename = await AddonManager.UI.save_file(filter_str="LL Mesh (*.llmesh)")
        if not filename:
            return
        with open(filename, "wb") as f:
            f.write(build_mesh_deformer(self.deform_joints))

    @handle_command()
    async def upload_deformer_as_mesh(self, _session: Session, region: ProxiedRegion):
        """Same as save_deformer_as_mesh, but uploads the mesh directly to SL."""

        mesh_bytes = build_mesh_deformer(self.deform_joints)
        try:
            # Send off mesh to calculate upload cost
            upload_token = await region.asset_uploader.initiate_mesh_upload("deformer", mesh_bytes)
        except Exception as e:
            show_message(e)
            raise

        if not await AddonManager.UI.confirm("Upload", f"Spend {upload_token.linden_cost}L on upload?"):
            return

        # Do the actual upload
        try:
            await region.asset_uploader.complete_upload(upload_token)
        except Exception as e:
            show_message(e)
            raise


addons = [DeformerAddon()]



from typing import *

from hippolyzer.lib.base.helpers import proxify
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.client.parcel_manager import ParcelManager
if TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion


class ProxyParcelManager(ParcelManager):
    def __init__(self, region: "ProxiedRegion"):
        super().__init__(proxify(region))
        # Handle ParcelProperties messages that we didn't specifically ask for
        self._region.message_handler.subscribe("ParcelProperties", self._handle_parcel_properties)

    def _handle_parcel_properties(self, msg: Message):
        self._process_parcel_properties(msg)
        return None



"""
Make all object textures Jeff Bezos

Helpful for migrating your region to AWS
"""

import copy
import ctypes
import random
import secrets

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


BEZOS_UUIDS = [UUID(x) for x in [
    "b8b8dcf9-758a-4539-ba63-793a01407236",
    "0010533a-cd41-44de-9a74-ab2125cbef8f",
]]


def _modify_crc(crc_tweak: int, crc_val: int):
    return ctypes.c_uint32(crc_val ^ crc_tweak).value


def _bezosify_te(local_id, parsed_te):
    parsed_te = copy.copy(parsed_te)
    parsed_te.Textures = {None: BEZOS_UUIDS[local_id % len(BEZOS_UUIDS)]}
    return parsed_te


class BezosifyAddon(BaseAddon):
    bezos_crc_xor: int = SessionProperty()

    def handle_session_init(self, session: Session):
        # We want CRCs that are stable for the duration of the session, but will
        # cause a cache miss for objects cached before this session. Generate a
        # random value to XOR all CRCs with
        self.bezos_crc_xor = secrets.randbits(32)

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.name == "ObjectUpdateCached":
            for block in message["ObjectData"]:
                # Cached only really has a CRC, this will force the cache miss.
                block["CRC"] = _modify_crc(self.bezos_crc_xor, block["CRC"])
        elif message.name == "ObjectUpdate":
            for block in message["ObjectData"]:
                block["CRC"] = _modify_crc(self.bezos_crc_xor, block["CRC"])
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue

                parsed_te = _bezosify_te(block["ID"], parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "ImprovedTerseObjectUpdate":
            for block in message["ObjectData"]:
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue
                update_data = block.deserialize_var("Data")

                parsed_te = _bezosify_te(update_data["ID"], parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "AvatarAppearance":
            for block in message["ObjectData"]:
                parsed_te = block.deserialize_var("TextureEntry")
                if not parsed_te:
                    continue
                # Need an integer ID to choose a bezos texture, just use
                # the last byte of the sender's UUID.
                sender_id = message["Sender"]["ID"]
                parsed_te = _bezosify_te(sender_id.bytes[-1], parsed_te)
                block.serialize_var("TextureEntry", parsed_te)
        elif message.name == "ObjectUpdateCompressed":
            for block in message["ObjectData"]:
                update_data = block.deserialize_var("Data")
                if not update_data:
                    continue
                update_data["CRC"] = _modify_crc(self.bezos_crc_xor, update_data["CRC"])
                if not update_data.get("TextureEntry"):
                    continue

                update_data["TextureEntry"] = _bezosify_te(
                    update_data["ID"],
                    update_data["TextureEntry"],
                )
                block.serialize_var("Data", update_data)
        elif message.name == "RegionHandshake":
            for field_name, val in message["RegionInfo"][0].items():
                if field_name.startswith("Terrain") and isinstance(val, UUID):
                    message["RegionInfo"][field_name] = random.choice(BEZOS_UUIDS)


addons = [BezosifyAddon()]
